#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import os

from proton import *

from . import common
from . import engine
from .common import pump, Skipped


def _sslCertpath(file):
    """ Return the full path to the certificate, keyfile, etc.
    """
    if os.name == "nt":
        if file.find("private-key") != -1:
            # The private key is not in a separate store
            return None
        # Substitute pkcs#12 equivalent for the CA/key store
        if file.endswith(".pem"):
            file = file[:-4] + ".p12"
    return os.path.join(os.path.dirname(__file__),
                        "ssl_db/%s" % file)


def _testSaslMech(self, mech, clientUser='user@proton', authUser='user@proton',
                  authzid=None, encrypted=False, authenticated=True):
    self.s1.allowed_mechs(mech)
    self.c1.open()
    self.c2.open()

    pump(self.t1, self.t2, 1024)

    if encrypted is not None:
        assert self.t2.encrypted == encrypted, encrypted
        assert self.t1.encrypted == encrypted, encrypted

    if authzid is None:
        authzid = authUser

    assert self.t2.authenticated == authenticated, authenticated
    assert self.t1.authenticated == authenticated, authenticated
    if authenticated:
        # Server
        assert self.t2.user == authUser
        assert self.s2.user == authUser
        assert self.s2.authorization == authzid, self.s2.authorization
        assert self.s2.mech == mech.strip()
        assert self.s2.outcome == SASL.OK, self.s2.outcome
        assert self.c2.state & Endpoint.LOCAL_ACTIVE and self.c2.state & Endpoint.REMOTE_ACTIVE,\
            "local_active=%s, remote_active=%s" % (
                self.c1.state & Endpoint.LOCAL_ACTIVE, self.c1.state & Endpoint.REMOTE_ACTIVE)
        # Client
        assert self.t1.user == clientUser
        assert self.s1.user == clientUser
        assert self.s1.mech == mech.strip()
        assert self.s1.outcome == SASL.OK, self.s1.outcome
        assert self.c1.state & Endpoint.LOCAL_ACTIVE and self.c1.state & Endpoint.REMOTE_ACTIVE,\
            "local_active=%s, remote_active=%s" % (
                self.c1.state & Endpoint.LOCAL_ACTIVE, self.c1.state & Endpoint.REMOTE_ACTIVE)
    else:
        # Server
        assert self.t2.user is None
        assert self.s2.user is None
        assert self.s2.authorization is None, self.s2.authorization
        assert self.s2.outcome != SASL.OK, self.s2.outcome
        # Client
        assert self.t1.user == clientUser
        assert self.s1.user == clientUser
        assert self.s1.outcome != SASL.OK, self.s1.outcome


class Test(common.Test):
    pass


def consumeAllOuput(t):
    stops = 0
    while stops < 1:
        out = t.peek(1024)
        l = len(out) if out else 0
        t.pop(l)
        if l <= 0:
            stops += 1


class SaslTest(Test):

    def setUp(self):
        self.t1 = Transport()
        self.s1 = SASL(self.t1)
        self.t2 = Transport(Transport.SERVER)
        self.t2.max_frame_size = 65536
        self.s2 = SASL(self.t2)

    def pump(self):
        pump(self.t1, self.t2, 1024)

    # We have to generate the client frames manually because proton does not
    # generate pipelined SASL and AMQP frames together
    def testIllegalProtocolLayering(self):
        # Server
        self.s2.allowed_mechs('ANONYMOUS')

        c2 = Connection()
        self.t2.bind(c2)

        assert self.s2.outcome is None

        # Push client bytes into server
        self.t2.push(
            # SASL
            b'AMQP\x03\x01\x00\x00'
            # @sasl-init(65) [mechanism=:ANONYMOUS, initial-response=b"anonymous@fuschia"]
            b'\x00\x00\x002\x02\x01\x00\x00\x00SA\xd0\x00\x00\x00"\x00\x00\x00\x02\xa3\x09ANONYMOUS\xa0\x11anonymous@fuschia'
            # SASL (again illegally)
            b'AMQP\x03\x01\x00\x00'
            # @sasl-init(65) [mechanism=:ANONYMOUS, initial-response=b"anonymous@fuschia"]
            b'\x00\x00\x002\x02\x01\x00\x00\x00SA\xd0\x00\x00\x00"\x00\x00\x00\x02\xa3\x09ANONYMOUS\xa0\x11anonymous@fuschia'
            # AMQP
            b'AMQP\x00\x01\x00\x00'
            # @open(16) [container-id="", channel-max=1234]
            b'\x00\x00\x00!\x02\x00\x00\x00\x00S\x10\xd0\x00\x00\x00\x11\x00\x00\x00\x0a\xa1\x00@@`\x04\xd2@@@@@@'
        )

        consumeAllOuput(self.t2)

        assert self.t2.condition
        assert self.t2.closed
        assert not c2.state & Endpoint.REMOTE_ACTIVE

    def testPipelinedClient(self):
        # Server
        self.s2.allowed_mechs('ANONYMOUS')

        c2 = Connection()
        self.t2.bind(c2)

        assert self.s2.outcome is None

        # Push client bytes into server
        self.t2.push(
            # SASL
            b'AMQP\x03\x01\x00\x00'
            # @sasl-init(65) [mechanism=:ANONYMOUS, initial-response=b"anonymous@fuschia"]
            b'\x00\x00\x002\x02\x01\x00\x00\x00SA\xd0\x00\x00\x00"\x00\x00\x00\x02\xa3\x09ANONYMOUS\xa0\x11anonymous@fuschia'
            # AMQP
            b'AMQP\x00\x01\x00\x00'
            # @open(16) [container-id="", channel-max=1234]
            b'\x00\x00\x00!\x02\x00\x00\x00\x00S\x10\xd0\x00\x00\x00\x11\x00\x00\x00\x0a\xa1\x00@@`\x04\xd2@@@@@@'
        )

        consumeAllOuput(self.t2)

        assert not self.t2.condition
        assert self.s2.outcome == SASL.OK
        assert c2.state & Endpoint.REMOTE_ACTIVE

    def testPipelinedServer(self):
        # Client
        self.s1.allowed_mechs('ANONYMOUS')

        c1 = Connection()
        self.t1.bind(c1)

        assert self.s1.outcome is None

        # Push server bytes into client
        # Commented out lines in this test are where the client input processing doesn't
        # run after output processing even though there is input waiting
        self.t1.push(
            # SASL
            b'AMQP\x03\x01\x00\x00'
            # @sasl-mechanisms(64) [sasl-server-mechanisms=@PN_SYMBOL[:ANONYMOUS]]
            b'\x00\x00\x00\x1c\x02\x01\x00\x00\x00S@\xc0\x0f\x01\xe0\x0c\x01\xa3\tANONYMOUS'
            # @sasl-outcome(68) [code=0]
            b'\x00\x00\x00\x10\x02\x01\x00\x00\x00SD\xc0\x03\x01P\x00'
            # AMQP
            b'AMQP\x00\x01\x00\x00'
            # @open(16) [container-id="", channel-max=1234]
            b'\x00\x00\x00!\x02\x00\x00\x00\x00S\x10\xd0\x00\x00\x00\x11\x00\x00\x00\x0a\xa1\x00@@`\x04\xd2@@@@@@'
        )

        consumeAllOuput(self.t1)

        assert self.s1.outcome == SASL.OK
        assert c1.state & Endpoint.REMOTE_ACTIVE

    def testPipelined2(self):
        out1 = self.t1.peek(1024)
        self.t1.pop(len(out1))
        self.t2.push(out1)

        self.s2.allowed_mechs('ANONYMOUS')
        c2 = Connection()
        c2.open()
        self.t2.bind(c2)

        out2 = self.t2.peek(1024)
        self.t2.pop(len(out2))
        self.t1.push(out2)

        out1 = self.t1.peek(1024)
        assert len(out1) > 0

    def testFracturedSASL(self):
        """ PROTON-235
        """
        assert self.s1.outcome is None

        # self.t1.trace(Transport.TRACE_FRM)

        out = self.t1.peek(1024)
        self.t1.pop(len(out))
        self.t1.push(b"AMQP\x03\x01\x00\x00")
        out = self.t1.peek(1024)
        self.t1.pop(len(out))
        self.t1.push(b"\x00\x00\x00")
        out = self.t1.peek(1024)
        self.t1.pop(len(out))

        self.t1.push(b"6\x02\x01\x00\x00\x00S@\xc0\x29\x01\xe0\x26\x04\xa3\x05PLAIN\x0aDIGEST-MD5\x09ANONYMOUS\x08CRAM-MD5")
        out = self.t1.peek(1024)
        self.t1.pop(len(out))
        self.t1.push(b"\x00\x00\x00\x10\x02\x01\x00\x00\x00SD\xc0\x03\x01P\x00")
        out = self.t1.peek(1024)
        self.t1.pop(len(out))
        while out:
            out = self.t1.peek(1024)
            self.t1.pop(len(out))

        assert self.s1.outcome == SASL.OK, self.s1.outcome

    def test_singleton(self):
        """Verify that only a single instance of SASL can exist per Transport"""
        transport = Transport()
        attr = object()
        sasl1 = SASL(transport)
        sasl1.my_attribute = attr
        sasl2 = transport.sasl()
        sasl3 = SASL(transport)
        assert sasl1 == sasl2
        assert sasl1 == sasl3
        assert sasl1.my_attribute == attr
        assert sasl2.my_attribute == attr
        assert sasl3.my_attribute == attr
        transport = Transport()
        sasl1 = transport.sasl()
        sasl1.my_attribute = attr
        sasl2 = SASL(transport)
        assert sasl1 == sasl2
        assert sasl1.my_attribute == attr
        assert sasl2.my_attribute == attr

    def testSaslSkipped(self):
        """Verify that the server (with SASL) correctly handles a client without SASL"""
        self.t1 = Transport()
        self.t2.require_auth(False)
        self.pump()
        assert self.s2.outcome is None
        assert self.t2.condition is None
        assert self.t2.authenticated == False
        assert self.s1.outcome is None
        assert self.t1.condition is None
        assert self.t1.authenticated == False

    def testSaslSkippedFail(self):
        """Verify that the server (with SASL) correctly handles a client without SASL"""
        self.t1 = Transport()
        self.t2.require_auth(True)
        self.pump()
        assert self.s2.outcome is None
        assert self.t2.condition is not None
        assert self.s1.outcome is None
        assert self.t1.condition is not None

    def testMechNotFound(self):
        self.c1 = Connection()
        self.c1.open()
        self.t1.bind(self.c1)
        self.s1.allowed_mechs('IMPOSSIBLE')

        self.pump()

        assert self.t2.authenticated == False
        assert self.t1.authenticated == False
        assert self.s1.outcome != SASL.OK
        assert self.s2.outcome != SASL.OK


class SASLMechTest(Test):
    def setUp(self):
        self.t1 = Transport()
        self.s1 = SASL(self.t1)
        self.t2 = Transport(Transport.SERVER)
        self.s2 = SASL(self.t2)

        self.c1 = Connection()
        self.c1.user = 'user@proton'
        self.c1.password = 'password'
        self.c1.hostname = 'localhost'

        self.c2 = Connection()

    def testANON(self):
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'ANONYMOUS', authUser='anonymous')

    def testCRAMMD5(self):
        common.ensureCanTestExtendedSASL()

        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'CRAM-MD5')

    def testDIGESTMD5(self):
        common.ensureCanTestExtendedSASL()

        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5')

    # PLAIN shouldn't work without encryption without special setting
    def testPLAINfail(self):
        common.ensureCanTestExtendedSASL()

        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'PLAIN', authenticated=False)

    # Client won't accept PLAIN even if offered by server without special setting
    def testPLAINClientFail(self):
        common.ensureCanTestExtendedSASL()

        self.s2.allow_insecure_mechs = True
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'PLAIN', authenticated=False)

    # PLAIN will only work if both ends are specially set up
    def testPLAIN(self):
        common.ensureCanTestExtendedSASL()

        self.s1.allow_insecure_mechs = True
        self.s2.allow_insecure_mechs = True
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'PLAIN')

# SCRAM not supported before Cyrus SASL 2.1.26
# so not universal and hence need a test for support
# to keep it in tests.
#  def testSCRAMSHA1(self):
#    common.ensureCanTestExtendedSASL()
#
#    self.t1.bind(self.c1)
#    self.t2.bind(self.c2)
#    _testSaslMech(self, 'SCRAM-SHA-1')


def _sslConnection(domain, transport, connection):
    transport.bind(connection)
    ssl = SSL(transport, domain, None)
    return connection


class SSLSASLTest(Test):
    def setUp(self):
        if not common.isSSLPresent():
            raise Skipped("No SSL libraries found.")

        self.server_domain = SSLDomain(SSLDomain.MODE_SERVER)
        self.client_domain = SSLDomain(SSLDomain.MODE_CLIENT)

        self.t1 = Transport()
        self.s1 = SASL(self.t1)
        self.t2 = Transport(Transport.SERVER)
        self.s2 = SASL(self.t2)

        self.c1 = Connection()
        self.c2 = Connection()

    def testSSLPlainSimple(self):
        if not SASL.extended():
            raise Skipped("Simple SASL server does not support PLAIN")
        common.ensureCanTestExtendedSASL()

        clientUser = 'user@proton'
        mech = 'PLAIN'

        self.c1.user = clientUser
        self.c1.password = 'password'
        self.c1.hostname = 'localhost'

        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        ssl1 = _sslConnection(self.client_domain, self.t1, self.c1)
        ssl2 = _sslConnection(self.server_domain, self.t2, self.c2)

        _testSaslMech(self, mech, encrypted=True)

    def testSSLPlainAuthzid(self):
        if not SASL.extended():
            raise Skipped("Simple SASL server does not support PLAIN")
        common.ensureCanTestExtendedSASL()

        clientUser = 'user@proton'
        authzid = 'user2'
        mech = 'PLAIN'

        self.c1.user = clientUser
        self.c1.authorization = authzid
        self.c1.password = 'password'
        self.c1.hostname = 'localhost'

        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        ssl1 = _sslConnection(self.client_domain, self.t1, self.c1)
        ssl2 = _sslConnection(self.server_domain, self.t2, self.c2)

        _testSaslMech(self, mech, authzid=authzid, encrypted=True)

    def testSSLPlainSimpleFail(self):
        if not SASL.extended():
            raise Skipped("Simple SASL server does not support PLAIN")
        common.ensureCanTestExtendedSASL()

        clientUser = 'usr@proton'
        mech = 'PLAIN'

        self.c1.user = clientUser
        self.c1.password = 'password'
        self.c1.hostname = 'localhost'

        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        ssl1 = _sslConnection(self.client_domain, self.t1, self.c1)
        ssl2 = _sslConnection(self.server_domain, self.t2, self.c2)

        _testSaslMech(self, mech, clientUser='usr@proton', encrypted=True, authenticated=False)

    def testSSLExternalSimple(self):
        if os.name == "nt":
            extUser = 'O=Client, CN=127.0.0.1'
        else:
            extUser = 'O=Client,CN=127.0.0.1'
        mech = 'EXTERNAL'

        self.server_domain.set_credentials(_sslCertpath("server-certificate.pem"),
                                           _sslCertpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(_sslCertpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   _sslCertpath("ca-certificate.pem"))
        self.client_domain.set_credentials(_sslCertpath("client-certificate.pem"),
                                           _sslCertpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_trusted_ca_db(_sslCertpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        ssl1 = _sslConnection(self.client_domain, self.t1, self.c1)
        ssl2 = _sslConnection(self.server_domain, self.t2, self.c2)

        _testSaslMech(self, mech, clientUser=None, authUser=extUser, encrypted=True)

    def testSSLExternalAuthzid(self):
        if os.name == "nt":
            extUser = 'O=Client, CN=127.0.0.1'
        else:
            extUser = 'O=Client,CN=127.0.0.1'
        mech = 'EXTERNAL'
        authzid = 'user_foo'

        self.c1.authorization = authzid

        self.server_domain.set_credentials(_sslCertpath("server-certificate.pem"),
                                           _sslCertpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(_sslCertpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   _sslCertpath("ca-certificate.pem"))
        self.client_domain.set_credentials(_sslCertpath("client-certificate.pem"),
                                           _sslCertpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_trusted_ca_db(_sslCertpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        ssl1 = _sslConnection(self.client_domain, self.t1, self.c1)
        ssl2 = _sslConnection(self.server_domain, self.t2, self.c2)

        _testSaslMech(self, mech, clientUser=None, authUser=extUser, authzid=authzid, encrypted=True)

    def testSSLExternalSimpleFail(self):
        mech = 'EXTERNAL'

        self.server_domain.set_credentials(_sslCertpath("server-certificate.pem"),
                                           _sslCertpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(_sslCertpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   _sslCertpath("ca-certificate.pem"))
        self.client_domain.set_trusted_ca_db(_sslCertpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        ssl1 = _sslConnection(self.client_domain, self.t1, self.c1)
        ssl2 = _sslConnection(self.server_domain, self.t2, self.c2)

        _testSaslMech(self, mech, clientUser=None, authUser=None, encrypted=None, authenticated=False)


class SASLEventTest(engine.CollectorTest):
    def setUp(self):
        engine.CollectorTest.setUp(self)
        self.t1 = Transport()
        self.s1 = SASL(self.t1)
        self.t2 = Transport(Transport.SERVER)
        self.s2 = SASL(self.t2)

        self.c1 = Connection()
        self.c1.user = 'user@proton'
        self.c1.password = 'password'
        self.c1.hostname = 'localhost'

        self.c2 = Connection()

        self.collector = Collector()

    def testNormalAuthenticationClient(self):
        common.ensureCanTestExtendedSASL()
        self.c1.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5')
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.CONNECTION_REMOTE_OPEN)

    def testNormalAuthenticationServer(self):
        common.ensureCanTestExtendedSASL()
        self.c2.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5')
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.CONNECTION_REMOTE_OPEN)

    def testFailedAuthenticationClient(self):
        common.ensureCanTestExtendedSASL()
        clientUser = "usr@proton"
        self.c1.user = clientUser
        self.c1.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5', clientUser=clientUser, authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_ERROR,
                    Event.TRANSPORT_TAIL_CLOSED,
                    Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testFailedAuthenticationServer(self):
        common.ensureCanTestExtendedSASL()
        clientUser = "usr@proton"
        self.c1.user = clientUser
        self.c2.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5', clientUser=clientUser, authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_ERROR,
                    Event.TRANSPORT_TAIL_CLOSED,
                    Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testNoMechClient(self):
        common.ensureCanTestExtendedSASL()
        self.c1.collect(self.collector)
        self.s2.allowed_mechs('IMPOSSIBLE')
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5', authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_ERROR,
                    Event.TRANSPORT_TAIL_CLOSED, Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testNoMechServer(self):
        common.ensureCanTestExtendedSASL()
        self.c2.collect(self.collector)
        self.s2.allowed_mechs('IMPOSSIBLE')
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'DIGEST-MD5', authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_TAIL_CLOSED,
                    Event.TRANSPORT_ERROR, Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testDisallowedMechClient(self):
        self.c1.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'IMPOSSIBLE', authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_ERROR,
                    Event.TRANSPORT_TAIL_CLOSED, Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testDisallowedMechServer(self):
        self.c2.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'IMPOSSIBLE', authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_TAIL_CLOSED,
                    Event.TRANSPORT_ERROR, Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testDisallowedPlainClient(self):
        self.c1.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'PLAIN', authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_ERROR,
                    Event.TRANSPORT_TAIL_CLOSED, Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)

    def testDisallowedPlainServer(self):
        self.c2.collect(self.collector)
        self.t1.bind(self.c1)
        self.t2.bind(self.c2)
        _testSaslMech(self, 'PLAIN', authenticated=False)
        self.expect(Event.CONNECTION_INIT, Event.CONNECTION_BOUND,
                    Event.CONNECTION_LOCAL_OPEN, Event.TRANSPORT,
                    Event.TRANSPORT_TAIL_CLOSED,
                    Event.TRANSPORT_ERROR, Event.TRANSPORT_HEAD_CLOSED, Event.TRANSPORT_CLOSED)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import time

from proton.reactor import Container, ApplicationEvent, EventInjector, Selector, Backoff
from proton.handlers import Handshaker, MessagingHandler
from proton import Handler, Url, symbol

from .common import Test, SkipTest, TestServer, free_tcp_port, free_tcp_ports, ensureCanTestExtendedSASL


class Barf(Exception):
    pass


class BarfOnInit:

    def on_reactor_init(self, event):
        raise Barf()

    def on_connection_init(self, event):
        raise Barf()

    def on_session_init(self, event):
        raise Barf()

    def on_link_init(self, event):
        raise Barf()


class BarfOnTask:

    def on_timer_task(self, event):
        raise Barf()


class BarfOnFinal:
    init = False

    def on_reactor_init(self, event):
        self.init = True

    def on_reactor_final(self, event):
        raise Barf()


class BarfOnFinalDerived(Handshaker):
    init = False

    def on_reactor_init(self, event):
        self.init = True

    def on_reactor_final(self, event):
        raise Barf()


class ExceptionTest(Test):

    def setUp(self):
        self.container = Container()

    def test_reactor_final(self):
        self.container.global_handler = BarfOnFinal()
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_global_set(self):
        self.container.global_handler = BarfOnInit()
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_global_add(self):
        self.container.global_handler.add(BarfOnInit())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_reactor_set(self):
        self.container.handler = BarfOnInit()
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_reactor_add(self):
        self.container.handler.add(BarfOnInit())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_connection(self):
        self.container.connection(BarfOnInit())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_connection_set(self):
        c = self.container.connection()
        c.handler = BarfOnInit()
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_connection_add(self):
        c = self.container.connection()
        c.handler = object()
        c.handler.add(BarfOnInit())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_session_set(self):
        c = self.container.connection()
        s = c.session()
        s.handler = BarfOnInit()
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_session_add(self):
        c = self.container.connection()
        s = c.session()
        s.handler = object()
        s.handler.add(BarfOnInit())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_link_set(self):
        c = self.container.connection()
        s = c.session()
        l = s.sender("xxx")
        l.handler = BarfOnInit()
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_link_add(self):
        c = self.container.connection()
        s = c.session()
        l = s.sender("xxx")
        l.handler = object()
        l.handler.add(BarfOnInit())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_schedule(self):
        self.container.schedule(0, BarfOnTask())
        try:
            self.container.run()
            assert False, "expected to barf"
        except Barf:
            pass

    def test_schedule_event(self):
        class Nothing:
            def __init__(self, p):
                self.parent = p

            results = []

            def on_timer_task(self, event):
                self.parent.triggered = True
                assert event.context == self.parent.task
                assert event.container == self.parent.container
        self.task = self.container.schedule(0, Nothing(self))
        self.container.run()
        assert self.triggered

    def test_schedule_many_nothings(self):
        class Nothing:
            results = []

            def on_timer_task(self, event):
                self.results.append(None)
        num = 12345
        for a in range(num):
            self.container.schedule(0, Nothing())
        self.container.run()
        assert len(Nothing.results) == num

    def test_schedule_many_nothing_refs(self):
        class Nothing:
            results = []

            def on_timer_task(self, event):
                self.results.append(None)
        num = 12345
        tasks = []
        for a in range(num):
            tasks.append(self.container.schedule(0, Nothing()))
        self.container.run()
        assert len(Nothing.results) == num

    def test_schedule_many_nothing_refs_cancel_before_run(self):
        class Nothing:
            results = []

            def on_timer_task(self, event):
                self.results.append(None)
        num = 12345
        tasks = []
        for a in range(num):
            tasks.append(self.container.schedule(0, Nothing()))
        for task in tasks:
            task.cancel()
        self.container.run()
        assert len(Nothing.results) == 0

    def test_schedule_cancel(self):
        barf = self.container.schedule(10, BarfOnTask())

        class CancelBarf:
            def __init__(self, barf):
                self.barf = barf

            def on_timer_task(self, event):
                self.barf.cancel()
                pass
        self.container.schedule(0, CancelBarf(barf))
        start = time.time()
        try:
            self.container.run()
            elapsed = time.time() - start
            assert elapsed < 10, "should have cancelled immediately, took %ss" % elapsed
        except Barf:
            assert False, "expected barf to be cancelled"

    def test_schedule_cancel_many(self):
        num = 12345
        barfs = set()
        for a in range(num):
            barf = self.container.schedule(10 * (a + 1), BarfOnTask())

            class CancelBarf:
                def __init__(self, barf):
                    self.barf = barf

                def on_timer_task(self, event):
                    self.barf.cancel()
                    barfs.discard(self.barf)
                    pass
            self.container.schedule(0, CancelBarf(barf))
            barfs.add(barf)
        start = time.time()
        try:
            self.container.run()
            elapsed = time.time() - start
            assert elapsed < num, "expected cancelled task to not delay the reactor by %s" % elapsed
            assert not barfs, "expected all barfs to be discarded"
        except Barf:
            assert False, "expected barf to be cancelled"


class ApplicationEventTest(Test):
    """Test application defined events and handlers."""

    class MyTestServer(TestServer):
        def __init__(self):
            super(ApplicationEventTest.MyTestServer, self).__init__()

    class MyHandler(Handler):
        def __init__(self, test):
            super(ApplicationEventTest.MyHandler, self).__init__()
            self._test = test

        def on_hello(self, event):
            # verify PROTON-1056
            self._test.hello_rcvd = str(event)

        def on_goodbye(self, event):
            self._test.goodbye_rcvd = str(event)

    def setUp(self):
        import os
        if not hasattr(os, 'pipe'):
            # KAG: seems like Jython doesn't have an os.pipe() method
            raise SkipTest()
        if os.name == "nt":
            # Correct implementation on Windows is complicated
            raise SkipTest("PROTON-1071")
        self.server = ApplicationEventTest.MyTestServer()
        self.server.reactor.handler.add(ApplicationEventTest.MyHandler(self))
        self.event_injector = EventInjector()
        self.hello_event = ApplicationEvent("hello")
        self.goodbye_event = ApplicationEvent("goodbye")
        self.server.reactor.selectable(self.event_injector)
        self.hello_rcvd = None
        self.goodbye_rcvd = None
        self.server.start()

    def tearDown(self):
        self.server.stop()

    def _wait_for(self, predicate, timeout=10.0):
        deadline = time.time() + timeout
        while time.time() < deadline:
            if predicate():
                break
            time.sleep(0.1)
        assert predicate()

    def test_application_events(self):
        self.event_injector.trigger(self.hello_event)
        self._wait_for(lambda: self.hello_rcvd is not None)
        self.event_injector.trigger(self.goodbye_event)
        self._wait_for(lambda: self.goodbye_rcvd is not None)


class AuthenticationTestHandler(MessagingHandler):
    def __init__(self):
        super(AuthenticationTestHandler, self).__init__()
        port = free_tcp_port()
        self.url = "localhost:%i" % port
        self.verified = False

    def on_start(self, event):
        self.listener = event.container.listen(self.url)

    def on_connection_opened(self, event):
        event.connection.close()

    def on_connection_opening(self, event):
        assert event.connection.transport.user == "user@proton"
        self.verified = True

    def on_connection_closed(self, event):
        event.connection.close()
        self.listener.close()

    def on_connection_error(self, event):
        event.connection.close()
        self.listener.close()


class ContainerTest(Test):
    """Test container subclass of reactor."""

    def test_event_has_container_attribute(self):
        ensureCanTestExtendedSASL()

        class TestHandler(MessagingHandler):
            def __init__(self):
                super(TestHandler, self).__init__()
                port = free_tcp_port()
                self.url = "localhost:%i" % port

            def on_start(self, event):
                self.listener = event.container.listen(self.url)

            def on_connection_closing(self, event):
                event.connection.close()
                self.listener.close()
        test_handler = TestHandler()
        container = Container(test_handler)

        class ConnectionHandler(MessagingHandler):
            def __init__(self):
                super(ConnectionHandler, self).__init__()

            def on_connection_opened(self, event):
                event.connection.close()
                assert event.container is event.reactor
                assert event.container is container
        container.connect(test_handler.url, handler=ConnectionHandler())
        container.run()

    def test_authentication_via_url(self):
        ensureCanTestExtendedSASL()
        test_handler = AuthenticationTestHandler()
        container = Container(test_handler)
        container.connect("%s:password@%s" % ("user%40proton", test_handler.url), reconnect=False)
        container.run()
        assert test_handler.verified

    def test_authentication_via_container_attributes(self):
        ensureCanTestExtendedSASL()
        test_handler = AuthenticationTestHandler()
        container = Container(test_handler)
        container.user = "user@proton"
        container.password = "password"
        container.connect(test_handler.url, reconnect=False)
        container.run()
        assert test_handler.verified

    def test_authentication_via_kwargs(self):
        ensureCanTestExtendedSASL()
        test_handler = AuthenticationTestHandler()
        container = Container(test_handler)
        container.connect(test_handler.url, user="user@proton", password="password", reconnect=False)
        container.run()
        assert test_handler.verified

    class _ServerHandler(MessagingHandler):
        def __init__(self, host):
            super(ContainerTest._ServerHandler, self).__init__()
            self.host = host
            self.port = free_tcp_port()
            self.client_addr = None
            self.peer_hostname = None

        def on_start(self, event):
            self.listener = event.container.listen("%s:%s" % (self.host, self.port))

        def on_connection_opened(self, event):
            self.client_addr = event.connected_address
            self.peer_hostname = event.connection.remote_hostname

        def on_connection_closing(self, event):
            event.connection.close()
            self.listener.close()

    class _ClientHandler(MessagingHandler):
        def __init__(self):
            super(ContainerTest._ClientHandler, self).__init__()
            self.server_addr = None
            self.errors = 0

        def on_connection_opened(self, event):
            self.server_addr = event.connected_address
            event.connection.close()

        def on_transport_error(self, event):
            self.errors += 1

    def test_numeric_hostname(self):
        ensureCanTestExtendedSASL()
        server_handler = ContainerTest._ServerHandler("127.0.0.1")
        client_handler = ContainerTest._ClientHandler()
        container = Container(server_handler)
        container.connect(url="127.0.0.1:%s" % (server_handler.port),
                          handler=client_handler)
        container.run()
        assert server_handler.client_addr
        assert client_handler.server_addr
        assert server_handler.peer_hostname == "127.0.0.1", server_handler.peer_hostname
        assert client_handler.server_addr.rsplit(':', 1)[1] == str(server_handler.port)

    def test_non_numeric_hostname(self):
        ensureCanTestExtendedSASL()
        server_handler = ContainerTest._ServerHandler("localhost")
        client_handler = ContainerTest._ClientHandler()
        container = Container(server_handler)
        container.connect(url="localhost:%s" % (server_handler.port),
                          handler=client_handler)
        container.run()
        assert server_handler.client_addr
        assert client_handler.server_addr
        assert server_handler.peer_hostname == "localhost", server_handler.peer_hostname
        assert client_handler.server_addr.rsplit(':', 1)[1] == str(server_handler.port)

    def test_virtual_host(self):
        ensureCanTestExtendedSASL()
        server_handler = ContainerTest._ServerHandler("localhost")
        container = Container(server_handler)
        conn = container.connect(url="localhost:%s" % (server_handler.port),
                                 handler=ContainerTest._ClientHandler(),
                                 virtual_host="a.b.c.org")
        container.run()
        assert server_handler.peer_hostname == "a.b.c.org", server_handler.peer_hostname

    def test_no_virtual_host(self):
        # explicitly setting an empty virtual host should prevent the hostname
        # field from being sent in the Open performative when using the
        # Python Container.
        server_handler = ContainerTest._ServerHandler("localhost")
        container = Container(server_handler)
        conn = container.connect(url="localhost:%s" % (server_handler.port),
                                 handler=ContainerTest._ClientHandler(),
                                 virtual_host="")
        container.run()
        assert server_handler.peer_hostname is None, server_handler.peer_hostname

    class _ReconnectServerHandler(MessagingHandler):
        def __init__(self, host, listen_on_error=True):
            super(ContainerTest._ReconnectServerHandler, self).__init__()
            self.host = host
            self.port = free_tcp_port()
            self.client_addr = None
            self.peer_hostname = None
            self.listen_on_error = listen_on_error

        def on_connection_opened(self, event):
            self.client_addr = event.connected_address
            self.peer_hostname = event.connection.remote_hostname
            self.listener.close()

        def on_connection_closing(self, event):
            event.connection.close()

        def listen(self, container):
            if self.listen_on_error:
                self.listener = container.listen("%s:%s" % (self.host, self.port))

    class _ReconnectClientHandler(MessagingHandler):
        def __init__(self, server_handler):
            super(ContainerTest._ReconnectClientHandler, self).__init__()
            self.connect_failed = False
            self.server_addr = None
            self.server_handler = server_handler

        def on_connection_opened(self, event):
            self.server_addr = event.connected_address
            event.connection.close()

        def on_transport_error(self, event):
            assert self.connect_failed == False
            self.connect_failed = True
            self.server_handler.listen(event.container)

    def test_failover(self):
        server_handler = ContainerTest._ServerHandler("localhost")
        client_handler = ContainerTest._ClientHandler()
        free_ports = free_tcp_ports(2)
        container = Container(server_handler)
        container.connect(urls=["localhost:%s" % (free_ports[0]), "localhost:%s" % (free_ports[1]),
                                "localhost:%s" % (server_handler.port)],
                          handler=client_handler)
        container.run()
        assert server_handler.peer_hostname == 'localhost', server_handler.peer_hostname
        assert client_handler.server_addr == Url(host='localhost', port=server_handler.port), client_handler.server_addr

    def test_failover_fail(self):
        client_handler = ContainerTest._ClientHandler()
        free_ports = free_tcp_ports(2)
        container = Container(client_handler)
        start = time.time()
        container.connect(urls=["localhost:%s" % (free_ports[0]), "localhost:%s" % (free_ports[1])],
                          reconnect=Backoff(max_tries=5),
                          handler=client_handler)
        container.run()
        end = time.time()
        assert client_handler.errors == 10
        # Total time for failure should be greater than but close to 3s
        # would like to have an upper bound of about 3.2 too - but loaded CI machines can take a loooong time!
        assert 3.0 < end - start, end - start
        assert client_handler.server_addr is None, client_handler.server_addr

    def test_failover_fail_custom_reconnect(self):
        client_handler = ContainerTest._ClientHandler()
        free_ports = free_tcp_ports(2)
        container = Container(client_handler)
        start = time.time()
        container.connect(urls=["localhost:%s" % (free_ports[0]), "localhost:%s" % (free_ports[1])],
                          reconnect=[0, 0.5, 1],
                          handler=client_handler)
        container.run()
        end = time.time()
        assert client_handler.errors == 6
        # Total time for failure should be greater than but close to 3s
        # would like to have an upper bound of about 3.2 too - but loaded CI machines can take a loooong time!
        assert 3.0 < end - start, end - start
        assert client_handler.server_addr is None, client_handler.server_addr

    def test_reconnect(self):
        server_handler = ContainerTest._ReconnectServerHandler("localhost", listen_on_error=True)
        client_handler = ContainerTest._ReconnectClientHandler(server_handler)
        container = Container(server_handler)
        container.connect(url="localhost:%s" % (server_handler.port),
                          handler=client_handler)
        container.run()
        assert server_handler.peer_hostname == 'localhost', server_handler.peer_hostname
        assert client_handler.connect_failed
        assert client_handler.server_addr == Url(host='localhost', port=server_handler.port), client_handler.server_addr

    def test_not_reconnecting(self):
        server_handler = ContainerTest._ReconnectServerHandler("localhost", listen_on_error=False)
        client_handler = ContainerTest._ReconnectClientHandler(server_handler)
        container = Container(server_handler)
        container.connect(url="localhost:%s" % (server_handler.port),
                          handler=client_handler, reconnect=False)
        container.run()
        assert server_handler.peer_hostname is None, server_handler.peer_hostname
        assert client_handler.connect_failed
        assert client_handler.server_addr is None, client_handler.server_addr


class SelectorTest(Test):
    """Test the Selector"""

    def test_unicode_selector(self):
        assert Selector(u"Hello").filter_set[symbol('selector')].value == u"Hello"

    def test_non_unicode_selector(self):
        assert Selector(b"Hello").filter_set[symbol('selector')].value == u"Hello"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import errno
import logging
import socket
import time
import weakref

from ._condition import Condition
from ._delivery import Delivery
from ._endpoints import Endpoint
from ._events import Event, Handler, _dispatch
from ._exceptions import ProtonException
from ._io import IO
from ._message import Message
from ._selectable import Selectable
from ._transport import Transport
from ._url import Url
from typing import Any, List, Optional, Tuple, Union, TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from ._delivery import DispositionType
    from ._reactor import Container, Transaction
    from ._endpoints import Sender, Receiver

log = logging.getLogger("proton")


class OutgoingMessageHandler(Handler):
    """
    A utility for simpler and more intuitive handling of delivery
    events related to outgoing i.e. sent messages.

    :param auto_settle: If ``True`` (default), automatically settle messages
        upon receiving a settled disposition for that delivery. Otherwise
        messages must be explicitly settled.
    :type auto_settle: ``bool``
    :param delegate: A client handler for the endpoint event
    """

    def __init__(self, auto_settle=True, delegate=None):
        self.auto_settle = auto_settle
        self.delegate = delegate

    def on_link_flow(self, event: Event):
        if event.link.is_sender and event.link.credit \
                and event.link.state & Endpoint.LOCAL_ACTIVE \
                and event.link.state & Endpoint.REMOTE_ACTIVE:
            self.on_sendable(event)

    def on_delivery(self, event: Event):
        dlv = event.delivery
        if dlv.link.is_sender and dlv.updated:
            if dlv.remote_state == Delivery.ACCEPTED:
                self.on_accepted(event)
            elif dlv.remote_state == Delivery.REJECTED:
                self.on_rejected(event)
            elif dlv.remote_state == Delivery.RELEASED or dlv.remote_state == Delivery.MODIFIED:
                self.on_released(event)
            if dlv.settled:
                self.on_settled(event)
                if self.auto_settle:
                    dlv.settle()

    def on_sendable(self, event: Event):
        """
        Called when the sender link has credit and messages can
        therefore be transferred.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_sendable', event)

    def on_accepted(self, event: Event):
        """
        Called when the remote peer accepts an outgoing message.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_accepted', event)

    def on_rejected(self, event: Event):
        """
        Called when the remote peer rejects an outgoing message.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_rejected', event)

    def on_released(self, event: Event):
        """
        Called when the remote peer releases an outgoing message. Note
        that this may be in response to either the ``RELEASE`` or ``MODIFIED``
        state as defined by the AMQP specification.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_released', event)

    def on_settled(self, event: Event):
        """
        Called when the remote peer has settled the outgoing
        message. This is the point at which it should never be
        retransmitted.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_settled', event)


def recv_msg(delivery: Delivery) -> Message:
    msg = Message()
    msg.decode(delivery.link.recv(delivery.pending))
    delivery.link.advance()
    return msg


class Reject(ProtonException):
    """
    An exception that indicates a message should be rejected.
    """
    pass


class Release(ProtonException):
    """
    An exception that indicates a message should be released.
    """
    pass


class Acking(object):
    """
    A class containing methods for handling received messages.
    """

    def accept(self, delivery: Delivery) -> None:
        """
        Accepts a received message.

        .. note:: This method cannot currently be used in combination
            with transactions. See :class:`proton.reactor.Transaction`
            for transactional methods.

        :param delivery: The message delivery tracking object
        """
        self.settle(delivery, Delivery.ACCEPTED)

    def reject(self, delivery: Delivery) -> None:
        """
        Rejects a received message that is considered invalid or
        unprocessable.

        .. note:: This method cannot currently be used in combination
            with transactions. See :class:`proton.reactor.Transaction`
            for transactional methods.

        :param delivery: The message delivery tracking object
        """
        self.settle(delivery, Delivery.REJECTED)

    def release(self, delivery: Delivery, delivered: bool = True) -> None:
        """
        Releases a received message, making it available at the source
        for any (other) interested receiver. The ``delivered``
        parameter indicates whether this should be considered a
        delivery attempt (and the delivery count updated) or not.

        .. note:: This method cannot currently be used in combination
            with transactions. See :class:`proton.reactor.Transaction`
            for transactional methods.

        :param delivery: The message delivery tracking object
        :param delivered: If ``True``, the message will be annotated
            with a delivery attempt (setting delivery flag
            :const:`proton.Delivery.MODIFIED`). Otherwise, the message
            will be returned without the annotation and released (setting
            delivery flag :const:`proton.Delivery.RELEASED`
        """
        if delivered:
            self.settle(delivery, Delivery.MODIFIED)
        else:
            self.settle(delivery, Delivery.RELEASED)

    def settle(self, delivery: Delivery, state: Optional['DispositionType'] = None) -> None:
        """
        Settles the message delivery, and optionally updating the
        delivery state.

        :param delivery: The message delivery tracking object
        :param state: The delivery state, or ``None`` if no update
            is to be performed.
        """
        if state:
            delivery.update(state)
        delivery.settle()


class IncomingMessageHandler(Handler, Acking):
    """
    A utility for simpler and more intuitive handling of delivery
    events related to incoming i.e. received messages.

    :param auto_accept: If ``True``, accept all messages (default). Otherwise
        messages must be individually accepted or rejected.
    :param delegate: A client handler for the endpoint event
    """

    def __init__(self, auto_accept: bool = True, delegate: Optional[Handler] = None) -> None:
        self.delegate = delegate
        self.auto_accept = auto_accept

    def on_delivery(self, event: Event) -> None:
        dlv = event.delivery
        if not dlv.link.is_receiver:
            return
        if dlv.aborted:
            self.on_aborted(event)
            dlv.settle()
        elif dlv.readable and not dlv.partial:
            event.message = recv_msg(dlv)
            if event.link.state & Endpoint.LOCAL_CLOSED:
                if self.auto_accept:
                    dlv.update(Delivery.RELEASED)
                    dlv.settle()
            else:
                try:
                    self.on_message(event)
                    if self.auto_accept:
                        dlv.update(Delivery.ACCEPTED)
                        dlv.settle()
                except Reject:
                    dlv.update(Delivery.REJECTED)
                    dlv.settle()
                except Release:
                    dlv.update(Delivery.MODIFIED)
                    dlv.settle()
        elif dlv.updated and dlv.settled:
            self.on_settled(event)

    def on_message(self, event: Event):
        """
        Called when a message is received. The message itself can be
        obtained as a property on the event. For the purpose of
        referring to this message in further actions (e.g. if
        explicitly accepting it, the ``delivery`` should be used, also
        obtainable via a property on the event.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_message', event)

    def on_settled(self, event: Event):
        """
        Callback for when a message delivery is settled by the remote peer.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_settled', event)

    def on_aborted(self, event: Event):
        """
        Callback for when a message delivery is aborted by the remote peer.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_aborted', event)


class EndpointStateHandler(Handler):
    """
    A utility that exposes 'endpoint' events - ie the open/close for
    links, sessions and connections in a more intuitive manner. A
    ``XXX_opened()`` method will be called when both local and remote peers
    have opened the link, session or connection. This can be used to
    confirm a locally initiated action for example. A ``XXX_opening()``
    method will be called when the remote peer has requested an open
    that was not initiated locally. By default this will simply open
    locally, which then triggers the ``XXX_opened()`` call. The same applies
    to close.

    :param peer_close_is_error: If ``True``, a peer endpoint closing will be
        treated as an error with an error callback. Otherwise (default), the
        normal callbacks for the closing will occur.
    :param delegate: A client handler for the endpoint event
    """

    def __init__(self, peer_close_is_error: bool = False, delegate: Optional[Handler] = None) -> None:
        self.delegate = delegate
        self.peer_close_is_error = peer_close_is_error

    @classmethod
    def is_local_open(cls, endpoint: Endpoint) -> bool:
        """
        Test if local ``endpoint`` is open (ie has state
        :const:`proton.Endpoint.LOCAL_ACTIVE`).

        :param endpoint: The local endpoint to be tested.
        :return: ``True`` if local endpoint is in state
            :const:`proton.Endpoint.LOCAL_ACTIVE`, ``False`` otherwise.
        """
        return bool(endpoint.state & Endpoint.LOCAL_ACTIVE)

    @classmethod
    def is_local_uninitialised(cls, endpoint: Endpoint) -> bool:
        """
        Test if local ``endpoint`` is uninitialised (ie has state
        :const:`proton.Endpoint.LOCAL_UNINIT`).

        :param endpoint: The local endpoint to be tested.
        :return: ``True`` if local endpoint is in state
            :const:`proton.Endpoint.LOCAL_UNINIT`, ``False`` otherwise.
        """
        return bool(endpoint.state & Endpoint.LOCAL_UNINIT)

    @classmethod
    def is_local_closed(cls, endpoint: Endpoint) -> bool:
        """
        Test if local ``endpoint`` is closed (ie has state
        :const:`proton.Endpoint.LOCAL_CLOSED`).

        :param endpoint: The local endpoint to be tested.
        :return: ``True`` if local endpoint is in state
            :const:`proton.Endpoint.LOCAL_CLOSED`, ``False`` otherwise.
        """
        return bool(endpoint.state & Endpoint.LOCAL_CLOSED)

    @classmethod
    def is_remote_open(cls, endpoint: Endpoint) -> bool:
        """
        Test if remote ``endpoint`` is open (ie has state
        :const:`proton.Endpoint.LOCAL_ACTIVE`).

        :param endpoint: The remote endpoint to be tested.
        :return: ``True`` if remote endpoint is in state
            :const:`proton.Endpoint.LOCAL_ACTIVE`, ``False`` otherwise.
        """
        return bool(endpoint.state & Endpoint.REMOTE_ACTIVE)

    @classmethod
    def is_remote_closed(cls, endpoint: Endpoint) -> bool:
        """
        Test if remote ``endpoint`` is closed (ie has state
        :const:`proton.Endpoint.REMOTE_CLOSED`).

        :param endpoint: The remote endpoint to be tested.
        :return: ``True`` if remote endpoint is in state
            :const:`proton.Endpoint.REMOTE_CLOSED`, ``False`` otherwise.
        """
        return bool(endpoint.state & Endpoint.REMOTE_CLOSED)

    @classmethod
    def print_error(cls, endpoint: Endpoint, endpoint_type: str) -> None:
        """
        Logs an error message related to an error condition at an endpoint.

        :param endpoint: The endpoint to be tested
        :param endpoint_type: The endpoint type as a string to be printed
            in the log message.
        """
        if endpoint.remote_condition:
            log.error(endpoint.remote_condition.description or endpoint.remote_condition.name)
        elif cls.is_local_open(endpoint) and cls.is_remote_closed(endpoint):
            log.error("%s closed by peer" % endpoint_type)

    def on_link_remote_close(self, event: Event) -> None:
        if event.link.remote_condition:
            self.on_link_error(event)
        elif self.is_local_closed(event.link):
            self.on_link_closed(event)
        else:
            self.on_link_closing(event)
        event.link.close()

    def on_session_remote_close(self, event: Event) -> None:
        if event.session.remote_condition:
            self.on_session_error(event)
        elif self.is_local_closed(event.session):
            self.on_session_closed(event)
        else:
            self.on_session_closing(event)
        event.session.close()

    def on_connection_remote_close(self, event: Event) -> None:
        if event.connection.remote_condition:
            if event.connection.remote_condition.name == "amqp:connection:forced":
                # Treat this the same as just having the transport closed by the peer without
                # sending any events. Allow reconnection to happen transparently.
                return
            self.on_connection_error(event)
        elif self.is_local_closed(event.connection):
            self.on_connection_closed(event)
        else:
            self.on_connection_closing(event)
        event.connection.close()

    def on_connection_local_open(self, event: Event) -> None:
        if self.is_remote_open(event.connection):
            self.on_connection_opened(event)

    def on_connection_remote_open(self, event: Event) -> None:
        if self.is_local_open(event.connection):
            self.on_connection_opened(event)
        elif self.is_local_uninitialised(event.connection):
            self.on_connection_opening(event)
            event.connection.open()

    def on_session_local_open(self, event: Event) -> None:
        if self.is_remote_open(event.session):
            self.on_session_opened(event)

    def on_session_remote_open(self, event: Event) -> None:
        if self.is_local_open(event.session):
            self.on_session_opened(event)
        elif self.is_local_uninitialised(event.session):
            self.on_session_opening(event)
            event.session.open()

    def on_link_local_open(self, event: Event) -> None:
        if self.is_remote_open(event.link):
            self.on_link_opened(event)

    def on_link_remote_open(self, event: Event) -> None:
        if self.is_local_open(event.link):
            self.on_link_opened(event)
        elif self.is_local_uninitialised(event.link):
            self.on_link_opening(event)
            event.link.open()

    def on_connection_opened(self, event: Event) -> None:
        """
        Callback for when both the local and remote endpoints of a
        connection have opened.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_connection_opened', event)

    def on_session_opened(self, event: Event) -> None:
        """
        Callback for when both the local and remote endpoints of a
        session have opened.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_session_opened', event)

    def on_link_opened(self, event: Event) -> None:
        """
        Callback for when both the local and remote endpoints of a
        link have opened.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_link_opened', event)

    def on_connection_opening(self, event: Event) -> None:
        """
        Callback for when a remote peer initiates the opening of
        a connection.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_connection_opening', event)

    def on_session_opening(self, event: Event) -> None:
        """
        Callback for when a remote peer initiates the opening of
        a session.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_session_opening', event)

    def on_link_opening(self, event: Event) -> None:
        """
        Callback for when a remote peer initiates the opening of
        a link.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_link_opening', event)

    def on_connection_error(self, event: Event) -> None:
        """
        Callback for when an initiated connection open fails.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_connection_error', event)
        else:
            self.print_error(event.connection, "connection")

    def on_session_error(self, event: Event) -> None:
        """
        Callback for when an initiated session open fails.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_session_error', event)
        else:
            self.print_error(event.session, "session")
            event.connection.close()

    def on_link_error(self, event: Event) -> None:
        """
        Callback for when an initiated link open fails.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_link_error', event)
        else:
            self.print_error(event.link, "link")
            event.connection.close()

    def on_connection_closed(self, event: Event) -> None:
        """
        Callback for when both the local and remote endpoints of a
        connection have closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_connection_closed', event)

    def on_session_closed(self, event: Event) -> None:
        """
        Callback for when both the local and remote endpoints of a
        session have closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_session_closed', event)

    def on_link_closed(self, event: Event) -> None:
        """
        Callback for when both the local and remote endpoints of a
        link have closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_link_closed', event)

    def on_connection_closing(self, event: Event) -> None:
        """
        Callback for when a remote peer initiates the closing of
        a connection.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_connection_closing', event)
        elif self.peer_close_is_error:
            self.on_connection_error(event)

    def on_session_closing(self, event: Event) -> None:
        """
        Callback for when a remote peer initiates the closing of
        a session.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_session_closing', event)
        elif self.peer_close_is_error:
            self.on_session_error(event)

    def on_link_closing(self, event: Event) -> None:
        """
        Callback for when a remote peer initiates the closing of
        a link.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None:
            _dispatch(self.delegate, 'on_link_closing', event)
        elif self.peer_close_is_error:
            self.on_link_error(event)

    def on_transport_tail_closed(self, event: Event) -> None:
        """
        Callback for when the transport tail has closed (ie no further input will
        be accepted by the transport).

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        self.on_transport_closed(event)

    def on_transport_closed(self, event: Event) -> None:
        """
        Callback for when the transport has closed - ie both the head (input) and
        tail (output) of the transport pipeline are closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if self.delegate is not None and event.connection and self.is_local_open(event.connection):
            _dispatch(self.delegate, 'on_disconnected', event)


class MessagingHandler(Handler, Acking):
    """
    A general purpose handler that makes the proton-c events somewhat
    simpler to deal with and/or avoids repetitive tasks for common use
    cases.

    :param prefetch: Initial flow credit for receiving messages, defaults to 10.
    :param auto_accept: If ``True``, accept all messages (default). Otherwise
        messages must be individually accepted or rejected.
    :param auto_settle: If ``True`` (default), automatically settle messages
        upon receiving a settled disposition for that delivery. Otherwise
        messages must be explicitly settled.
    :param peer_close_is_error: If ``True``, a peer endpoint closing will be
        treated as an error with an error callback. Otherwise (default), the
        normal callbacks for the closing will occur.
    """

    def __init__(
            self,
            prefetch: int = 10,
            auto_accept: bool = True,
            auto_settle: bool = True,
            peer_close_is_error: bool = False
    ) -> None:
        self.handlers = []
        if prefetch:
            self.handlers.append(FlowController(prefetch))
        self.handlers.append(EndpointStateHandler(peer_close_is_error, weakref.proxy(self)))
        self.handlers.append(IncomingMessageHandler(auto_accept, weakref.proxy(self)))
        self.handlers.append(OutgoingMessageHandler(auto_settle, weakref.proxy(self)))
        self.fatal_conditions = ["amqp:unauthorized-access"]

    def on_transport_error(self, event: Event) -> None:
        """
        Called when some error is encountered with the transport over
        which the AMQP connection is to be established. This includes
        authentication errors as well as socket errors.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if event.transport.condition:
            if event.transport.condition.info:
                log.error("%s: %s: %s" % (
                    event.transport.condition.name, event.transport.condition.description,
                    event.transport.condition.info))
            else:
                log.error("%s: %s" % (event.transport.condition.name, event.transport.condition.description))
            if event.transport.condition.name in self.fatal_conditions:
                event.connection.close()
        else:
            logging.error("Unspecified transport error")

    def on_connection_error(self, event: Event) -> None:
        """
        Called when the peer closes the connection with an error condition.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        :type event: :class:`proton.Event`
        """
        EndpointStateHandler.print_error(event.connection, "connection")

    def on_session_error(self, event: Event) -> None:
        """
        Called when the peer closes the session with an error condition.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        EndpointStateHandler.print_error(event.session, "session")
        event.connection.close()

    def on_link_error(self, event: Event) -> None:
        """
        Called when the peer closes the link with an error condition.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        EndpointStateHandler.print_error(event.link, "link")
        event.connection.close()

    def on_reactor_init(self, event: Event) -> None:
        """
        Called when the event loop - the reactor - starts.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        if hasattr(event.reactor, 'subclass'):
            setattr(event, event.reactor.subclass.__name__.lower(), event.reactor)
        self.on_start(event)

    def on_start(self, event: Event) -> None:
        """
        Called when the event loop starts. (Just an alias for on_reactor_init)

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_connection_closed(self, event: Event) -> None:
        """
        Called when the connection is closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_session_closed(self, event: Event) -> None:
        """
        Called when the session is closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_link_closed(self, event: Event) -> None:
        """
        Called when the link is closed.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_connection_closing(self, event: Event) -> None:
        """
        Called when the peer initiates the closing of the connection.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_session_closing(self, event: Event) -> None:
        """
        Called when the peer initiates the closing of the session.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_link_closing(self, event: Event) -> None:
        """
        Called when the peer initiates the closing of the link.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_disconnected(self, event: Event) -> None:
        """
        Called when the socket is disconnected.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_sendable(self, event: Event) -> None:
        """
        Called when the sender link has credit and messages can
        therefore be transferred.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_accepted(self, event: Event) -> None:
        """
        Called when the remote peer accepts an outgoing message.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_rejected(self, event: Event) -> None:
        """
        Called when the remote peer rejects an outgoing message.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_released(self, event: Event) -> None:
        """
        Called when the remote peer releases an outgoing message. Note
        that this may be in response to either the RELEASE or MODIFIED
        state as defined by the AMQP specification.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_settled(self, event: Event) -> None:
        """
        Called when the remote peer has settled the outgoing
        message. This is the point at which it should never be
        retransmitted.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_message(self, event: Event) -> None:
        """
        Called when a message is received. The message itself can be
        obtained as a property on the event. For the purpose of
        referring to this message in further actions (e.g. if
        explicitly accepting it, the ``delivery`` should be used, also
        obtainable via a property on the event.

        :param event: The underlying event object. Use this to obtain further
            information on the event. In particular, the message itself may
            be obtained by accessing ``event.message``.
        """
        pass


class TransactionHandler(object):
    """
    The interface for transaction handlers - ie objects that want to
    be notified of state changes related to a transaction.
    """

    def on_transaction_declared(self, event: Event) -> None:
        """
        Called when a local transaction is declared.

        :param event: The underlying event object. Use this to obtain further
            information on the event. In particular, the :class:`proton.reactor.Transaction`
            object may be obtained by accessing ``event.transaction``.
        """
        pass

    def on_transaction_committed(self, event: Event) -> None:
        """
        Called when a local transaction is discharged successfully
        (committed).

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_transaction_aborted(self, event: Event) -> None:
        """
        Called when a local transaction is discharged unsuccessfully
        (aborted).

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_transaction_declare_failed(self, event: Event) -> None:
        """
        Called when a local transaction declare fails.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass

    def on_transaction_commit_failed(self, event: Event) -> None:
        """
        Called when the commit of a local transaction fails.

        :param event: The underlying event object. Use this to obtain further
            information on the event.
        """
        pass


class TransactionalClientHandler(MessagingHandler, TransactionHandler):
    """
    An extension to the MessagingHandler for applications using
    transactions. This handler provides all of the callbacks found
    in :class:`MessagingHandler` and :class:`TransactionHandler`,
    and provides a convenience method :meth:`accept` for performing
    a transactional acceptance of received messages.

    :param prefetch: Initial flow credit for receiving messages, defaults to 10.
    :param auto_accept: If ``True``, accept all messages (default). Otherwise
        messages must be individually accepted or rejected.
    :param auto_settle: If ``True`` (default), automatically settle messages
        upon receiving a settled disposition for that delivery. Otherwise
        messages must be explicitly settled.
    :param peer_close_is_error: If ``True``, a peer endpoint closing will be
        treated as an error with an error callback. Otherwise (default), the
        normal callbacks for the closing will occur.
    """

    def __init__(
            self,
            prefetch: int = 10,
            auto_accept: bool = False,
            auto_settle: bool = True,
            peer_close_is_error: bool = False
    ) -> None:
        super(TransactionalClientHandler, self).__init__(prefetch, auto_accept, auto_settle, peer_close_is_error)

    def accept(self, delivery: Delivery, transaction: Optional['Transaction'] = None):
        """
        A convenience method for accepting a received message as part of a
        transaction. If no transaction object is supplied, a regular
        non-transactional acceptance will be performed.

        :param delivery: Delivery tracking object for received message.
        :param transaction: Transaction tracking object which is required if
            the message is being accepted under the transaction. If ``None`` (default),
            then a normal non-transactional accept occurs.
        """
        if transaction:
            transaction.accept(delivery)
        else:
            super(TransactionalClientHandler, self).accept(delivery)


class FlowController(Handler):
    def __init__(self, window: int = 1024) -> None:
        self._window = window
        self._drained = 0

    def on_link_local_open(self, event: Event) -> None:
        self._flow(event.link)

    def on_link_remote_open(self, event: Event) -> None:
        self._flow(event.link)

    def on_link_flow(self, event: Event) -> None:
        self._flow(event.link)

    def on_delivery(self, event: Event) -> None:
        self._flow(event.link)

    def _flow(self, link: Union['Sender', 'Receiver']) -> None:
        if link.is_receiver:
            self._drained += link.drained()
            if self._drained == 0:
                delta = self._window - link.credit
                link.flow(delta)


class Handshaker(Handler):

    @staticmethod
    def on_connection_remote_open(event: Event) -> None:
        conn = event.connection
        if conn.state & Endpoint.LOCAL_UNINIT:
            conn.open()

    @staticmethod
    def on_session_remote_open(event: Event) -> None:
        ssn = event.session
        if ssn.state & Endpoint.LOCAL_UNINIT:
            ssn.open()

    @staticmethod
    def on_link_remote_open(event: Event) -> None:
        link = event.link
        if link.state & Endpoint.LOCAL_UNINIT:
            link.source.copy(link.remote_source)
            link.target.copy(link.remote_target)
            link.open()

    @staticmethod
    def on_connection_remote_close(event: Event) -> None:
        conn = event.connection
        if not conn.state & Endpoint.LOCAL_CLOSED:
            conn.close()

    @staticmethod
    def on_session_remote_close(event: Event) -> None:
        ssn = event.session
        if not ssn.state & Endpoint.LOCAL_CLOSED:
            ssn.close()

    @staticmethod
    def on_link_remote_close(event: Event) -> None:
        link = event.link
        if not link.state & Endpoint.LOCAL_CLOSED:
            link.close()


# Back compatibility definitions
CFlowController = FlowController
CHandshaker = Handshaker


class PythonIO:

    def __init__(self) -> None:
        self.selectables = []
        self.delegate = IOHandler()

    def on_unhandled(self, method: str, event: Event) -> None:
        event.dispatch(self.delegate)

    def on_selectable_init(self, event: Event) -> None:
        self.selectables.append(event.context)

    def on_selectable_updated(self, event: Event) -> None:
        pass

    def on_selectable_final(self, event: Event) -> None:
        sel = event.context
        if sel.is_terminal:
            self.selectables.remove(sel)
            sel.close()

    def on_reactor_quiesced(self, event: Event) -> None:
        reactor = event.reactor
        # check if we are still quiesced, other handlers of
        # on_reactor_quiesced could have produced events to process
        if not reactor.quiesced:
            return

        reading = []
        writing = []
        deadline = None
        for sel in self.selectables:
            if sel.reading:
                reading.append(sel)
            if sel.writing:
                writing.append(sel)
            if sel.deadline:
                if deadline is None:
                    deadline = sel.deadline
                else:
                    deadline = min(sel.deadline, deadline)

        if deadline is not None:
            timeout = deadline - time.time()
        else:
            timeout = reactor.timeout
        if timeout < 0:
            timeout = 0
        timeout = min(timeout, reactor.timeout)
        readable, writable, _ = IO.select(reading, writing, [], timeout)

        now = reactor.mark()

        for s in readable:
            s.readable()
        for s in writable:
            s.writable()
        for s in self.selectables:
            if s.deadline and now > s.deadline:
                s.expired()

        reactor.yield_()


# For C style IO handler need to implement Selector
class IOHandler(Handler):

    def __init__(self) -> None:
        self._selector = IO.Selector()

    def on_selectable_init(self, event: Event) -> None:
        s = event.selectable
        self._selector.add(s)
        s._reactor._selectables += 1

    def on_selectable_updated(self, event: Event) -> None:
        s = event.selectable
        self._selector.update(s)

    def on_selectable_final(self, event: Event) -> None:
        s = event.selectable
        self._selector.remove(s)
        s._reactor._selectables -= 1
        s.close()

    def on_reactor_quiesced(self, event: Event) -> None:
        r = event.reactor

        if not r.quiesced:
            return

        d = r.timer_deadline
        readable, writable, expired = self._selector.select(r.timeout)

        now = r.mark()

        for s in readable:
            s.readable()
        for s in writable:
            s.writable()
        for s in expired:
            s.expired()

        r.yield_()

    def on_selectable_readable(self, event: Event) -> None:
        s = event.selectable
        t = s._transport

        # If we're an acceptor we can't have a transport
        # and we don't want to do anything here in any case
        if not t:
            return

        capacity = t.capacity()
        if capacity > 0:
            try:
                b = s.recv(capacity)
                if len(b) > 0:
                    n = t.push(b)
                else:
                    # EOF handling
                    self.on_selectable_error(event)
            except socket.error as e:
                # TODO: What's the error handling to be here?
                log.error("Couldn't recv: %r" % e)
                t.close_tail()

        # Always update as we may have gone to not reading or from
        # not writing to writing when processing the incoming bytes
        r = s._reactor
        self.update(t, s, r.now)

    def on_selectable_writable(self, event: Event) -> None:
        s = event.selectable
        t = s._transport

        # If we're an acceptor we can't have a transport
        # and we don't want to do anything here in any case
        if not t:
            return

        pending = t.pending()
        if pending > 0:

            try:
                n = s.send(t.peek(pending))
                t.pop(n)
            except socket.error as e:
                log.error("Couldn't send: %r" % e)
                # TODO: Error? or actually an exception
                t.close_head()

        newpending = t.pending()
        if newpending != pending:
            r = s._reactor
            self.update(t, s, r.now)

    def on_selectable_error(self, event: Event) -> None:
        s = event.selectable
        t = s._transport

        t.close_head()
        t.close_tail()
        s.terminate()
        s._transport = None
        t._selectable = None
        s.update()

    def on_selectable_expired(self, event: Event) -> None:
        s = event.selectable
        t = s._transport
        r = s._reactor

        self.update(t, s, r.now)

    def on_connection_local_open(self, event: Event) -> None:
        c = event.connection
        if not c.state & Endpoint.REMOTE_UNINIT:
            return

        t = Transport()
        # It seems perverse, but the C code ignores bind errors too!
        # and this is required or you get errors because Connector() has already
        # bound the transport and connection!
        t.bind_nothrow(c)

    def on_connection_bound(self, event: Event) -> None:
        c = event.connection
        t = event.transport

        reactor = c._reactor

        # link the new transport to its reactor:
        t._reactor = reactor

        if c._acceptor:
            # this connection was created by the acceptor.  There is already a
            # socket assigned to this connection.  Nothing needs to be done.
            return

        url = c.url or Url(c.hostname)
        url.defaults()

        host = url.host
        port = int(url.port)

        if not c.user:
            user = url.username
            if user:
                c.user = user
            password = url.password
            if password:
                c.password = password

        addrs = socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM)

        # Try first possible address
        log.debug("Connect trying first transport address: %s", addrs[0])
        sock = IO.connect(addrs[0])

        # At this point we need to arrange to be called back when the socket is writable
        ConnectSelectable(sock, reactor, addrs[1:], t, self)

        # TODO: Don't understand why we need this now - how can we get PN_TRANSPORT until the connection succeeds?
        t._selectable = None

    @staticmethod
    def update(transport: Transport, selectable: Selectable, now: float) -> None:
        try:
            capacity = transport.capacity()
            selectable.reading = capacity > 0
        except:
            if transport.closed:
                selectable.terminate()
                selectable._transport = None
                transport._selectable = None
        try:
            pending = transport.pending()
            selectable.writing = pending > 0
        except:
            if transport.closed:
                selectable.terminate()
                selectable._transport = None
                transport._selectable = None
        selectable.deadline = transport.tick(now)
        selectable.update()

    def on_transport(self, event: Event) -> None:
        t = event.transport
        r = t._reactor
        s = t._selectable
        if s and not s.is_terminal:
            self.update(t, s, r.now)

    def on_transport_closed(self, event: Event) -> None:
        t = event.transport
        r = t._reactor
        s = t._selectable
        if s and not s.is_terminal:
            s.terminate()
            s._transport = None
            t._selectable = None
            r.update(s)
        t.unbind()


class ConnectSelectable(Selectable):
    def __init__(
            self,
            sock: socket.socket,
            reactor: 'Container',
            addrs: List[Any],
            transport: Transport,
            iohandler: IOHandler
    ) -> None:
        super(ConnectSelectable, self).__init__(sock, reactor)
        self.writing = True
        self._addrs = addrs
        self._transport = transport
        self._iohandler = iohandler
        transport._connect_selectable = self

    def readable(self) -> None:
        pass

    def writable(self) -> None:
        e = self._delegate.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        t = self._transport
        t._connect_selectable = None

        # Always cleanup this ConnectSelectable: either we failed or created a new one
        # Do it first to ensure the socket gets deregistered before being registered again
        # in the case of connecting
        self.terminate()
        self._transport = None
        self.update()

        if e == 0:
            log.debug("Connection succeeded")

            # Disassociate from the socket (which will be passed on)
            self.release()

            s = self._reactor.selectable(delegate=self._delegate)
            s._transport = t
            t._selectable = s
            self._iohandler.update(t, s, t._reactor.now)

            return
        elif e == errno.ECONNREFUSED:
            if len(self._addrs) > 0:
                log.debug("Connection refused: trying next transport address: %s", self._addrs[0])

                sock = IO.connect(self._addrs[0])
                # New ConnectSelectable for the new socket with rest of addresses
                ConnectSelectable(sock, self._reactor, self._addrs[1:], t, self._iohandler)
                return
            else:
                log.debug("Connection refused, but tried all transport addresses")
                t.condition = Condition("proton.pythonio", "Connection refused to all addresses")
        else:
            log.error("Couldn't connect: %s", e)
            t.condition = Condition("proton.pythonio", "Connection error: %s" % e)

        t.close_tail()
        t.close_head()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import subprocess
import time
import unittest

import test_subprocess

# Check if we can run prlimit to control resources
try:
    assert subprocess.check_call(["prlimit"], stdout=open(os.devnull, 'w')) == 0, 'prlimit is present, but broken'
    prlimit_available = True
except OSError:
    prlimit_available = False


class PRLimitedBroker(test_subprocess.Server):
    def __init__(self, fdlimit, *args, **kwargs):
        super(PRLimitedBroker, self).__init__(
            ['prlimit', '-n{0:d}:'.format(fdlimit), "broker", "", "0"],  # `-n 256:` sets only soft limit to 256
            stdout=subprocess.PIPE, universal_newlines=True, *args, **kwargs)
        self.fdlimit = fdlimit


class FdLimitTest(unittest.TestCase):
    devnull = open(os.devnull, 'w')

    @classmethod
    def tearDownClass(cls):
        if cls.devnull:
            cls.devnull.close()

    @unittest.skipUnless(prlimit_available, "prlimit not available")
    def test_fd_limit_broker(self):
        """Check behaviour when running out of file descriptors on accept"""
        # Not too many FDs but not too few either, some are used for system purposes.
        fdlimit = 256
        with PRLimitedBroker(fdlimit, kill_me=True) as b:
            receivers = []

            # Start enough receivers to use all FDs
            # NOTE: broker does not log a file descriptor related error at any point in the test, only
            #  PN_TRANSPORT_CLOSED: amqp:connection:framing-error: connection aborted
            #  PN_TRANSPORT_CLOSED: proton:io: Connection reset by peer - disconnected :5672 (connection aborted)
            for i in range(fdlimit):
                receiver = subprocess.Popen(["receive", "", b.port, str(i)], stdout=self.devnull, stderr=subprocess.STDOUT)
                receivers.append(receiver)
            # Allow these subprocesses time to establish ahead of the upcoming test sender.
            time.sleep(1)

            # All FDs are now in use, new send should not succeed.  May fail by hanging (epoll) or by
            # immediate failure (libuv). But poll() should never be 0.
            sender = subprocess.Popen(["send", "", b.port, "x"],
                                       stdout=self.devnull, stderr=subprocess.STDOUT)
            time.sleep(1)  # polling for None immediately would always succeed, regardless whether send hangs or not
            self.assertNotEqual(sender.poll(), 0)

            # Kill receivers to free up FDs
            for r in receivers:
                r.kill()
            for r in receivers:
                r.wait()

            # Sender completes on its own
            sender.wait()

            # Additional send/receive should succeed now
            self.assertIn("10 messages sent", test_subprocess.check_output(
                ["send", "", b.port], universal_newlines=True))
            self.assertIn("10 messages received", test_subprocess.check_output(
                ["receive", "", b.port], universal_newlines=True))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import optparse
from proton import Message
from proton.handlers import MessagingHandler
from proton.reactor import Container


class Client(MessagingHandler):
    def __init__(self, url, requests):
        super(Client, self).__init__()
        self.url = url
        self.requests = requests

    def on_start(self, event):
        self.sender = event.container.create_sender(self.url)
        self.receiver = event.container.create_receiver(self.sender.connection, None, dynamic=True)

    def next_request(self):
        if self.receiver.remote_source.address:
            req = Message(reply_to=self.receiver.remote_source.address, body=self.requests[0])
            self.sender.send(req)

    def on_link_opened(self, event):
        if event.receiver == self.receiver:
            self.next_request()

    def on_message(self, event):
        print("%s => %s" % (self.requests.pop(0), event.message.body))
        if self.requests:
            self.next_request()
        else:
            event.connection.close()


REQUESTS = ["Twas brillig, and the slithy toves",
            "Did gire and gymble in the wabe.",
            "All mimsy were the borogroves,",
            "And the mome raths outgrabe."]

parser = optparse.OptionParser(usage="usage: %prog [options]",
                               description="Send requests to the supplied address and print responses.")
parser.add_option("-a", "--address", default="localhost:5672/examples",
                  help="address to which messages are sent (default %default)")
opts, args = parser.parse_args()

Container(Client(opts.address, args or REQUESTS)).run()



from __future__ import absolute_import
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import os

from .common import Test, Skipped, free_tcp_ports, \
    MessengerReceiverC, MessengerSenderC, \
    ReactorReceiverC, ReactorSenderC, \
    isSSLPresent

#
# Tests that run the apps
#


class AppTests(Test):

    def __init__(self, *args):
        Test.__init__(self, *args)
        self.is_valgrind = False

    def default(self, name, value, **kwargs):
        if self.is_valgrind:
            default = kwargs.get("valgrind", value)
        else:
            default = value
        return Test.default(self, name, default, **kwargs)

    @property
    def iterations(self):
        return int(self.default("iterations", 2, fast=1, valgrind=2))

    @property
    def send_count(self):
        return int(self.default("send_count", 17, fast=1, valgrind=2))

    @property
    def target_count(self):
        return int(self.default("target_count", 5, fast=1, valgrind=2))

    @property
    def send_batch(self):
        return int(self.default("send_batch", 7, fast=1, valgrind=2))

    @property
    def forward_count(self):
        return int(self.default("forward_count", 5, fast=1, valgrind=2))

    @property
    def port_count(self):
        return int(self.default("port_count", 3, fast=1, valgrind=2))

    @property
    def sender_count(self):
        return int(self.default("sender_count", 3, fast=1, valgrind=2))

    def valgrind_test(self):
        self.is_valgrind = True

    def setUp(self):
        self.senders = []
        self.receivers = []

    def tearDown(self):
        pass

    def _do_test(self, iterations=1):
        verbose = self.verbose

        for R in self.receivers:
            R.start(verbose)

        for j in range(iterations):
            for S in self.senders:
                S.start(verbose)

            for S in self.senders:
                S.wait()
                #print("SENDER OUTPUT:")
                #print( S.stdout() )
                assert S.status() == 0, ("Command '%s' failed status=%d: '%s' '%s'"
                                         % (str(S.cmdline()),
                                            S.status(),
                                            S.stdout(),
                                            S.stderr()))

        for R in self.receivers:
            R.wait()
            #print("RECEIVER OUTPUT")
            #print( R.stdout() )
            assert R.status() == 0, ("Command '%s' failed status=%d: '%s' '%s'"
                                     % (str(R.cmdline()),
                                        R.status(),
                                        R.stdout(),
                                        R.stderr()))

#
# Traffic passing tests based on the Messenger apps
#


class MessengerTests(AppTests):

    _timeout = 60

    def _ssl_check(self):
        if not isSSLPresent():
            raise Skipped("No SSL libraries found.")
        if os.name == "nt":
            raise Skipped("Windows SChannel lacks anonymous cipher support.")

    def __init__(self, *args):
        AppTests.__init__(self, *args)

    def _do_oneway_test(self, receiver, sender, domain="amqp"):
        """ Send N messages to a receiver.
        Parameters:
        iterations - repeat the senders this many times
        target_count = # of targets to send to.
        send_count = # messages sent to each target
        """
        iterations = self.iterations
        send_count = self.send_count
        target_count = self.target_count

        send_total = send_count * target_count
        receive_total = send_total * iterations

        port = free_tcp_ports()[0]

        receiver.subscriptions = ["%s://~0.0.0.0:%s" % (domain, port)]
        receiver.receive_count = receive_total
        receiver.timeout = MessengerTests._timeout
        self.receivers.append(receiver)

        sender.targets = ["%s://0.0.0.0:%s/X%d" % (domain, port, j) for j in range(target_count)]
        sender.send_count = send_total
        sender.timeout = MessengerTests._timeout
        self.senders.append(sender)

        self._do_test(iterations)

    def _do_echo_test(self, receiver, sender, domain="amqp"):
        """ Send N messages to a receiver, which responds to each.
        Parameters:
        iterations - repeat the senders this many times
        target_count - # targets to send to
        send_count = # messages sent to each target
        send_batch - wait for replies after this many messages sent
        """
        iterations = self.iterations
        send_count = self.send_count
        target_count = self.target_count
        send_batch = self.send_batch

        send_total = send_count * target_count
        receive_total = send_total * iterations

        port = free_tcp_ports()[0]

        receiver.subscriptions = ["%s://~0.0.0.0:%s" % (domain, port)]
        receiver.receive_count = receive_total
        receiver.send_reply = True
        receiver.timeout = MessengerTests._timeout
        self.receivers.append(receiver)

        sender.targets = ["%s://0.0.0.0:%s/%dY" % (domain, port, j) for j in range(target_count)]
        sender.send_count = send_total
        sender.get_reply = True
        sender.send_batch = send_batch
        sender.timeout = MessengerTests._timeout
        self.senders.append(sender)

        self._do_test(iterations)

    # Removed messenger "relay" tests. The test start-up is faulty:
    # msgr-recv prints it's -X ready message when it starts to open a
    # connection but it does not wait for the remote open. The relay
    # tests depends on mapping a container name from an incoming
    # connection. They can fail under if the sender starts before the
    # connection is complete (esp. valgrind with SSL connections) We
    # could fix the tests but since messenger is deprecated it does
    # not seem worthwhile.

    def _do_star_topology_test(self, r_factory, s_factory, domain="amqp"):
        """
        A star-like topology, with a central receiver at the hub, and senders at
        the spokes.  Each sender will connect to each of the ports the receiver is
        listening on.  Each sender will then create N links per each connection.
        Each sender will send X messages per link, waiting for a response.
        Parameters:
        iterations - repeat the senders this many times
        port_count - # of ports the receiver will listen on.  Each sender connects
                     to all ports.
        sender_count - # of senders
        target_count - # of targets per connection
        send_count - # of messages sent to each target
        send_batch - # of messages to send before waiting for response
        """
        iterations = self.iterations
        port_count = self.port_count
        sender_count = self.sender_count
        target_count = self.target_count
        send_count = self.send_count
        send_batch = self.send_batch

        send_total = port_count * target_count * send_count
        receive_total = send_total * sender_count * iterations

        ports = free_tcp_ports(port_count)

        receiver = r_factory()
        receiver.subscriptions = ["%s://~0.0.0.0:%s" % (domain, port) for port in ports]
        receiver.receive_count = receive_total
        receiver.send_reply = True
        receiver.timeout = MessengerTests._timeout
        self.receivers.append(receiver)

        for i in range(sender_count):
            sender = s_factory()
            sender.targets = ["%s://0.0.0.0:%s/%d" % (domain, port, j) for port in ports for j in range(target_count)]
            sender.send_count = send_total
            sender.send_batch = send_batch
            sender.get_reply = True
            sender.timeout = MessengerTests._timeout
            self.senders.append(sender)

        self._do_test(iterations)

    def test_oneway_C(self):
        self._do_oneway_test(MessengerReceiverC(), MessengerSenderC())

    def test_oneway_C_SSL(self):
        self._ssl_check()
        self._do_oneway_test(MessengerReceiverC(), MessengerSenderC(), "amqps")

    def test_echo_C(self):
        self._do_echo_test(MessengerReceiverC(), MessengerSenderC())

    def test_echo_C_SSL(self):
        self._ssl_check()
        self._do_echo_test(MessengerReceiverC(), MessengerSenderC(), "amqps")

    def test_star_topology_C(self):
        self._do_star_topology_test(MessengerReceiverC, MessengerSenderC)

    def test_star_topology_C_SSL(self):
        self._ssl_check()
        self._do_star_topology_test(MessengerReceiverC, MessengerSenderC, "amqps")

    def test_oneway_reactor(self):
        self._do_oneway_test(ReactorReceiverC(), ReactorSenderC())



# -----------------------------------------------------------------------------
#  Copyright (C) PyZMQ Developers
#  Distributed under the terms of the Modified BSD License.
#
#  This bundling code is largely adapted from pyzmq-static's get.sh by
#  Brandon Craig-Rhodes, which is itself BSD licensed.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Logging (adapted from h5py: http://h5py.googlecode.com)
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
#  This log code is largely adapted from pyzmq's code
#  PyZMQ Developers, which is itself Modified BSD licensed.
# -----------------------------------------------------------------------------


import os
import sys
import logging


logger = logging.getLogger()
if os.environ.get('DEBUG'):
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stderr))


def debug(msg):
    logger.debug(msg)


def info(msg):
    logger.info(msg)


def fatal(msg, code=1):
    logger.error("Fatal: " + msg)
    exit(code)


def warn(msg):
    logger.error("Warning: " + msg)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from cproton import PN_ACCEPTED, PN_MODIFIED, PN_RECEIVED, PN_REJECTED, PN_RELEASED, pn_delivery_abort, \
    pn_delivery_aborted, pn_delivery_attachments, pn_delivery_link, pn_delivery_local, pn_delivery_local_state, \
    pn_delivery_partial, pn_delivery_pending, pn_delivery_readable, pn_delivery_remote, pn_delivery_remote_state, \
    pn_delivery_settle, pn_delivery_settled, pn_delivery_tag, pn_delivery_update, pn_delivery_updated, \
    pn_delivery_writable, pn_disposition_annotations, pn_disposition_condition, pn_disposition_data, \
    pn_disposition_get_section_number, pn_disposition_get_section_offset, pn_disposition_is_failed, \
    pn_disposition_is_undeliverable, pn_disposition_set_failed, pn_disposition_set_section_number, \
    pn_disposition_set_section_offset, pn_disposition_set_undeliverable, pn_disposition_type, pn_work_next

from ._condition import cond2obj, obj2cond
from ._data import dat2obj, obj2dat
from ._wrapper import Wrapper

from typing import Dict, List, Optional, Type, Union, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ._condition import Condition
    from ._data import PythonAMQPData, symbol
    from ._endpoints import Receiver, Sender  # circular import
    from ._reactor import Connection, Session, Transport


class NamedInt(int):
    values: Dict[int, 'DispositionType'] = {}

    def __new__(cls: Type['DispositionType'], i: int, name: str) -> 'DispositionType':
        ni = super(NamedInt, cls).__new__(cls, i)
        cls.values[i] = ni
        return ni

    def __init__(self, i: int, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get(cls, i: int) -> Union[int, 'DispositionType']:
        return cls.values.get(i, i)


class DispositionType(NamedInt):
    values = {}


class Disposition(object):
    """
    A delivery state.

    Dispositions record the current state or final outcome of a
    transfer. Every delivery contains both a local and remote
    disposition. The local disposition holds the local state of the
    delivery, and the remote disposition holds the last known remote
    state of the delivery.
    """

    RECEIVED = DispositionType(PN_RECEIVED, "RECEIVED")
    """
    A non terminal state indicating how much (if any) message data
    has been received for a delivery.
    """

    ACCEPTED = DispositionType(PN_ACCEPTED, "ACCEPTED")
    """
    A terminal state indicating that the delivery was successfully
    processed. Once in this state there will be no further state
    changes prior to the delivery being settled.
    """

    REJECTED = DispositionType(PN_REJECTED, "REJECTED")
    """
    A terminal state indicating that the delivery could not be
    processed due to some error condition. Once in this state
    there will be no further state changes prior to the delivery
    being settled.
    """

    RELEASED = DispositionType(PN_RELEASED, "RELEASED")
    """
    A terminal state indicating that the delivery is being
    returned to the sender. Once in this state there will be no
    further state changes prior to the delivery being settled.
    """

    MODIFIED = DispositionType(PN_MODIFIED, "MODIFIED")
    """
    A terminal state indicating that the delivery is being
    returned to the sender and should be annotated by the
    sender prior to further delivery attempts. Once in this
    state there will be no further state changes prior to the
    delivery being settled.
    """

    def __init__(self, impl, local):
        self._impl = impl
        self.local = local
        self._data = None
        self._condition = None
        self._annotations = None

    @property
    def type(self) -> Union[int, DispositionType]:
        """
        Get the type of this disposition object.

        Defined values are:

        * :const:`RECEIVED`
        * :const:`ACCEPTED`
        * :const:`REJECTED`
        * :const:`RELEASED`
        * :const:`MODIFIED`
        """
        return DispositionType.get(pn_disposition_type(self._impl))

    @property
    def section_number(self) -> int:
        """The section number associated with a disposition."""
        return pn_disposition_get_section_number(self._impl)

    @section_number.setter
    def section_number(self, n: int) -> None:
        pn_disposition_set_section_number(self._impl, n)

    @property
    def section_offset(self) -> int:
        """The section offset associated with a disposition."""
        return pn_disposition_get_section_offset(self._impl)

    @section_offset.setter
    def section_offset(self, n: int) -> None:
        pn_disposition_set_section_offset(self._impl, n)

    @property
    def failed(self) -> bool:
        """The failed flag for this disposition."""
        return pn_disposition_is_failed(self._impl)

    @failed.setter
    def failed(self, b: bool) -> None:
        pn_disposition_set_failed(self._impl, b)

    @property
    def undeliverable(self) -> bool:
        """The undeliverable flag for this disposition."""
        return pn_disposition_is_undeliverable(self._impl)

    @undeliverable.setter
    def undeliverable(self, b: bool) -> None:
        pn_disposition_set_undeliverable(self._impl, b)

    @property
    def data(self) -> Optional[List[int]]:
        """Access the disposition as a :class:`Data` object.

        Dispositions are an extension point in the AMQP protocol. The
        disposition interface provides setters/getters for those
        dispositions that are predefined by the specification, however
        access to the raw disposition data is provided so that other
        dispositions can be used.

        The :class:`Data` object returned by this operation is valid until
        the parent delivery is settled.
        """
        if self.local:
            return self._data
        else:
            return dat2obj(pn_disposition_data(self._impl))

    @data.setter
    def data(self, obj: List[int]) -> None:
        if self.local:
            self._data = obj
        else:
            raise AttributeError("data attribute is read-only")

    @property
    def annotations(self) -> Optional[Dict['symbol', 'PythonAMQPData']]:
        """The annotations associated with a disposition.

        The :class:`Data` object retrieved by this operation may be modified
        prior to updating a delivery. When a delivery is updated, the
        annotations described by the :class:`Data` are reported to the peer
        if applicable to the current delivery state, e.g. states such as
        :const:`MODIFIED`. The :class:`Data` must be empty or contain a symbol
        keyed map.

        The :class:`Data` object returned by this operation is valid until
        the parent delivery is settled.
        """
        if self.local:
            return self._annotations
        else:
            return dat2obj(pn_disposition_annotations(self._impl))

    @annotations.setter
    def annotations(self, obj: Dict[str, 'PythonAMQPData']) -> None:
        if self.local:
            self._annotations = obj
        else:
            raise AttributeError("annotations attribute is read-only")

    @property
    def condition(self) -> Optional['Condition']:
        """The condition object associated with a disposition.

        The :class:`Condition` object retrieved by this operation may be
        modified prior to updating a delivery. When a delivery is updated,
        the condition described by the disposition is reported to the peer
        if applicable to the current delivery state, e.g. states such as
        :const:`REJECTED`.
        """
        if self.local:
            return self._condition
        else:
            return cond2obj(pn_disposition_condition(self._impl))

    @condition.setter
    def condition(self, obj: 'Condition') -> None:
        if self.local:
            self._condition = obj
        else:
            raise AttributeError("condition attribute is read-only")


class Delivery(Wrapper):
    """
    Tracks and/or records the delivery of a message over a link.
    """

    RECEIVED = Disposition.RECEIVED
    """
    A non terminal state indicating how much (if any) message data
    has been received for a delivery.
    """

    ACCEPTED = Disposition.ACCEPTED
    """
    A terminal state indicating that the delivery was successfully
    processed. Once in this state there will be no further state
    changes prior to the delivery being settled.
    """

    REJECTED = Disposition.REJECTED
    """
    A terminal state indicating that the delivery could not be
    processed due to some error condition. Once in this state
    there will be no further state changes prior to the delivery
    being settled.
    """

    RELEASED = Disposition.RELEASED
    """
    A terminal state indicating that the delivery is being
    returned to the sender. Once in this state there will be no
    further state changes prior to the delivery being settled.
    """

    MODIFIED = Disposition.MODIFIED
    """
    A terminal state indicating that the delivery is being
    returned to the sender and should be annotated by the
    sender prior to further delivery attempts. Once in this
    state there will be no further state changes prior to the
    delivery being settled.
    """

    @staticmethod
    def wrap(impl):
        if impl is None:
            return None
        else:
            return Delivery(impl)

    def __init__(self, impl):
        Wrapper.__init__(self, impl, pn_delivery_attachments)

    def _init(self) -> None:
        self.local = Disposition(pn_delivery_local(self._impl), True)
        self.remote = Disposition(pn_delivery_remote(self._impl), False)

    @property
    def tag(self) -> str:
        """
        The identifier for the delivery.
        """
        return pn_delivery_tag(self._impl)

    @property
    def writable(self) -> bool:
        """
        ``True`` for an outgoing delivery to which data can now be written,
        ``False`` otherwise..
        """
        return pn_delivery_writable(self._impl)

    @property
    def readable(self) -> bool:
        """
        ``True`` for an incoming delivery that has data to read,
        ``False`` otherwise..
        """
        return pn_delivery_readable(self._impl)

    @property
    def updated(self) -> bool:
        """
        ``True`` if the state of the delivery has been updated
        (e.g. it has been settled and/or accepted, rejected etc),
        ``False`` otherwise.
        """
        return pn_delivery_updated(self._impl)

    def update(self, state: Union[int, DispositionType]) -> None:
        """
        Set the local state of the delivery e.g. :const:`ACCEPTED`,
        :const:`REJECTED`, :const:`RELEASED`.

        :param state: State of delivery
        """
        obj2dat(self.local._data, pn_disposition_data(self.local._impl))
        obj2dat(self.local._annotations, pn_disposition_annotations(self.local._impl))
        obj2cond(self.local._condition, pn_disposition_condition(self.local._impl))
        pn_delivery_update(self._impl, state)

    @property
    def pending(self) -> int:
        """
        The amount of pending message data for a delivery.
        """
        return pn_delivery_pending(self._impl)

    @property
    def partial(self) -> bool:
        """
        ``True`` for an incoming delivery if not all the data is
        yet available, ``False`` otherwise.
        """
        return pn_delivery_partial(self._impl)

    @property
    def local_state(self) -> DispositionType:
        """A local state of the delivery."""
        return DispositionType.get(pn_delivery_local_state(self._impl))

    @property
    def remote_state(self) -> Union[int, DispositionType]:
        """A remote state of the delivery as indicated by the remote peer."""
        return DispositionType.get(pn_delivery_remote_state(self._impl))

    @property
    def settled(self) -> bool:
        """
        ``True`` if the delivery has been settled by the remote peer,
        ``False`` otherwise.
        """
        return pn_delivery_settled(self._impl)

    def settle(self) -> None:
        """
        Settles the delivery locally. This indicates the application
        considers the delivery complete and does not wish to receive any
        further events about it. Every delivery should be settled locally.
        """
        pn_delivery_settle(self._impl)

    @property
    def aborted(self) -> bool:
        """
        ``True`` if the delivery has been aborted, ``False`` otherwise.
        """
        return pn_delivery_aborted(self._impl)

    def abort(self) -> None:
        """
        Aborts the delivery.  This indicates the application wishes to
        invalidate any data that may have already been sent on this delivery.
        The delivery cannot be aborted after it has been completely delivered.
        """
        pn_delivery_abort(self._impl)

    @property
    def work_next(self) -> Optional['Delivery']:
        """Deprecated: use on_message(), on_accepted(), on_rejected(),
        on_released(), and on_settled() instead.

        The next :class:`Delivery` on the connection that has pending
        operations.
        """
        return Delivery.wrap(pn_work_next(self._impl))

    @property
    def link(self) -> Union['Receiver', 'Sender']:
        """
        The :class:`Link` on which the delivery was sent or received.
        """
        from . import _endpoints
        return _endpoints.Link.wrap(pn_delivery_link(self._impl))

    @property
    def session(self) -> 'Session':
        """
        The :class:`Session` over which the delivery was sent or received.
        """
        return self.link.session

    @property
    def connection(self) -> 'Connection':
        """
        The :class:`Connection` over which the delivery was sent or received.
        """
        return self.session.connection

    @property
    def transport(self) -> 'Transport':
        """
        The :class:`Transport` bound to the :class:`Connection` over which
        the delivery was sent or received.
        """
        return self.connection.transport



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# TODO: summarize, test harness preconditions (e.g. broker is alive)

import optparse
import os
import struct
import sys
import time
import traceback
import types
import cgi
from fnmatch import fnmatchcase as match
from logging import getLogger, StreamHandler, Formatter, Filter, \
    WARN, DEBUG, ERROR, INFO

from .common import SkipTest

if sys.version_info[0] == 3:
    CLASS_TYPES = (type,)
else:
    CLASS_TYPES = (type, types.ClassType)

levels = {
    "DEBUG": DEBUG,
    "INFO": INFO,
    "WARN": WARN,
    "ERROR": ERROR
}

sorted_levels = [k for k, v in sorted(levels.items(), key=lambda kv: kv[1])]

parser = optparse.OptionParser(usage="usage: %prog [options] PATTERN ...",
                               description="Run tests matching the specified PATTERNs.")
parser.add_option("-l", "--list", action="store_true", default=False,
                  help="list tests instead of executing them")
parser.add_option("-f", "--log-file", metavar="FILE", help="log output to FILE")
parser.add_option("-v", "--log-level", metavar="LEVEL", default="WARN",
                  help="only display log messages of LEVEL or higher severity: "
                  "%s (default %%default)" % ", ".join(sorted_levels))
parser.add_option("-c", "--log-category", metavar="CATEGORY", action="append",
                  dest="log_categories", default=[],
                  help="log only categories matching CATEGORY pattern")
parser.add_option("-m", "--module", action="append", default=[],
                  dest="modules", help="add module to test search path")
parser.add_option("-i", "--ignore", action="append", default=[],
                  help="ignore tests matching IGNORE pattern")
parser.add_option("-I", "--ignore-file", metavar="IFILE", action="append",
                  default=[],
                  help="ignore tests matching patterns in IFILE")
parser.add_option("-H", "--halt-on-error", action="store_true", default=False,
                  dest="hoe", help="halt if an error is encountered")
parser.add_option("-t", "--time", action="store_true", default=False,
                  help="report timing information on test run")
parser.add_option("-D", "--define", metavar="DEFINE", dest="defines",
                  action="append", default=[], help="define test parameters")
parser.add_option("-x", "--xml", metavar="XML", dest="xml",
                  help="write test results in Junit style xml suitable for use by CI tools etc")
parser.add_option("-a", "--always-colorize", action="store_true", dest="always_colorize", default=False,
                  help="always colorize the test results rather than relying on terminal tty detection. Useful when invoked from Jython/Maven.")
parser.add_option("-n", metavar="count", dest="count", type=int, default=1,
                  help="run the tests <count> times")
parser.add_option("-b", "--bare", action="store_true", default=False,
                  help="Run bare, i.e. don't capture stack traces. This is useful under Jython as " +
                  "captured stack traces do not include the Java portion of the stack," +
                  "whereas non captured stack traces do.")
parser.add_option("-j", "--javatrace", action="store_true", default=False,
                  help="Show the full java stack trace. This disables heuristics to eliminate the " +
                  "jython portion of java stack traces.")


class Config:

    def __init__(self):
        self.defines = {}
        self.log_file = None
        self.log_level = WARN
        self.log_categories = []


opts, args = parser.parse_args()

# Known bad tests, skipped unless specifically matched on the command line
known_bad = ["proton_tests.messenger.*"]

includes = []
excludes = ["*__*__"]
config = Config()
list_only = opts.list
for d in opts.defines:
    try:
        idx = d.index("=")
        name = d[:idx]
        value = d[idx + 1:]
        config.defines[name] = value
    except ValueError:
        config.defines[d] = None
config.log_file = opts.log_file
config.log_level = levels[opts.log_level.upper()]
config.log_categories = opts.log_categories
excludes.extend([v.strip() for v in opts.ignore])
for v in opts.ignore_file:
    f = open(v)
    for line in f:
        line = line.strip()
        if line.startswith("#"):
            continue
        excludes.append(line)
    f.close()

for a in args:
    includes.append(a.strip())


def is_ignored(path):
    for p in excludes:
        if match(path, p):
            return True
    return False


def is_included(path):
    if is_ignored(path):
        return False
    for p in includes:
        if match(path, p):
            return True
    for p in known_bad:
        if match(path, p):
            return False
    return not includes           # If includes is empty, include everything


def is_smart():
    return sys.stdout.isatty() and os.environ.get("TERM", "dumb") != "dumb"


try:
    import fcntl
    import termios

    def width():
        if is_smart():
            s = struct.pack("HHHH", 0, 0, 0, 0)
            fd_stdout = sys.stdout.fileno()
            x = fcntl.ioctl(fd_stdout, termios.TIOCGWINSZ, s)
            rows, cols, xpx, ypx = struct.unpack("HHHH", x)
            return cols
        else:
            try:
                return int(os.environ.get("COLUMNS", "80"))
            except ValueError:
                return 80

    WIDTH = width()

    def resize(sig, frm):
        global WIDTH
        WIDTH = width()

    import signal
    signal.signal(signal.SIGWINCH, resize)

except ImportError:
    WIDTH = 80


def vt100_attrs(*attrs):
    return "\x1B[%sm" % ";".join(map(str, attrs))


vt100_reset = vt100_attrs(0)

KEYWORDS = {"pass": (32,),
            "skip": (33,),
            "fail": (31,),
            "start": (34,),
            "total": (34,),
            "ignored": (33,),
            "selected": (34,),
            "elapsed": (34,),
            "average": (34,)}


def colorize_word(word, text=None):
    if text is None:
        text = word
    return colorize(text, *KEYWORDS.get(word, ()))


def colorize(text, *attrs):
    if attrs and (is_smart() or opts.always_colorize):
        return "%s%s%s" % (vt100_attrs(*attrs), text, vt100_reset)
    else:
        return text


def indent(text):
    lines = text.split("\n")
    return "  %s" % "\n  ".join(lines)

# Write a 'minimal' Junit xml style report file suitable for use by CI tools such as Jenkins.


class JunitXmlStyleReporter:

    def __init__(self, file):
        self.f = open(file, "w")

    def begin(self):
        self.f.write('<?xml version="1.0" encoding="UTF-8" ?>\n')
        self.f.write('<testsuite name="pythontests">\n')

    def report(self, name, result):
        parts = name.split(".")
        method = parts[-1]
        module = '.'.join(parts[0:-1])
        self.f.write('<testcase classname="%s" name="%s" time="%f">\n' % (module, method, result.time))
        if result.failed:
            escaped_type = cgi.escape(str(result.exception_type))
            escaped_message = cgi.escape(str(result.exception_message))
            self.f.write('<failure type="%s" message="%s">\n' % (escaped_type, escaped_message))
            self.f.write('<![CDATA[\n')
            self.f.write(result.formatted_exception_trace)
            self.f.write(']]>\n')
            self.f.write('</failure>\n')
        if result.skipped:
            self.f.write('<skipped/>\n')
        self.f.write('</testcase>\n')

    def end(self):
        self.f.write('</testsuite>\n')
        self.f.close()


class Interceptor:

    def __init__(self):
        self.newline = False
        self.indent = False
        self.passthrough = True
        self.dirty = False
        self.last = None

    def begin(self):
        self.newline = True
        self.indent = True
        self.passthrough = False
        self.dirty = False
        self.last = None

    def reset(self):
        self.newline = False
        self.indent = False
        self.passthrough = True


class StreamWrapper:

    def __init__(self, interceptor, stream, prefix="  "):
        self.interceptor = interceptor
        self.stream = stream
        self.prefix = prefix

    def fileno(self):
        return self.stream.fileno()

    def isatty(self):
        return self.stream.isatty()

    def write(self, s):
        if self.interceptor.passthrough:
            self.stream.write(s)
            return

        if s:
            self.interceptor.dirty = True

        if self.interceptor.newline:
            self.interceptor.newline = False
            self.stream.write(" %s\n" % colorize_word("start"))
            self.interceptor.indent = True
        if self.interceptor.indent:
            self.stream.write(self.prefix)
        if s.endswith("\n"):
            s = s.replace("\n", "\n%s" % self.prefix)[:-2]
            self.interceptor.indent = True
        else:
            s = s.replace("\n", "\n%s" % self.prefix)
            self.interceptor.indent = False
        self.stream.write(s)

        if s:
            self.interceptor.last = s[-1]

    def flush(self):
        self.stream.flush()


interceptor = Interceptor()

out_wrp = StreamWrapper(interceptor, sys.stdout)
err_wrp = StreamWrapper(interceptor, sys.stderr)

out = sys.stdout
err = sys.stderr
sys.stdout = out_wrp
sys.stderr = err_wrp


class PatternFilter(Filter):

    def __init__(self, *patterns):
        Filter.__init__(self, patterns)
        self.patterns = patterns

    def filter(self, record):
        if not self.patterns:
            return True
        for p in self.patterns:
            if match(record.name, p):
                return True
        return False


root = getLogger()
handler = StreamHandler(sys.stdout)
filter = PatternFilter(*config.log_categories)
handler.addFilter(filter)
handler.setFormatter(Formatter("%(asctime)s %(levelname)s %(message)s"))
root.addHandler(handler)
root.setLevel(WARN)

log = getLogger("proton.test")

PASS = "pass"
SKIP = "skip"
FAIL = "fail"


class Runner:

    def __init__(self):
        self.exception = None
        self.exception_phase_name = None
        self.skip = False

    def passed(self):
        return not self.exception

    def skipped(self):
        return self.skip

    def failed(self):
        return self.exception and not self.skip

    def halt(self):
        """determines if the overall test execution should be allowed to continue to the next phase"""
        return self.exception or self.skip

    def run(self, phase_name, phase):
        """invokes a test-phase method (which can be the test method itself or a setUp/tearDown
           method).  If the method raises an exception the exception is examined to see if the
           exception should be classified as a 'skipped' test"""
        # we don't try to catch exceptions for jython because currently a
        # jython bug will prevent the java portion of the stack being
        # stored with the exception info in the sys module
        if opts.bare:
            phase()
        else:
            try:
                phase()
            except KeyboardInterrupt:
                raise
            except:
                self.exception_phase_name = phase_name
                self.exception = sys.exc_info()
                exception_type = self.exception[0]
                self.skip = getattr(exception_type, "skipped", False) or issubclass(exception_type, SkipTest)

    def status(self):
        if self.passed():
            return PASS
        elif self.skipped():
            return SKIP
        elif self.failed():
            return FAIL
        else:
            return None

    def get_formatted_exception_trace(self):
        if self.exception:
            if self.skip:
                # format skipped tests without a traceback
                output = indent("".join(traceback.format_exception_only(*self.exception[:2]))).rstrip()
            else:
                output = "Error during %s:" % self.exception_phase_name
                lines = traceback.format_exception(*self.exception)
                val = self.exception[1]
                reflect = False
                if val and hasattr(val, "getStackTrace"):
                    jlines = []
                    for frame in val.getStackTrace():
                        if reflect and frame.getClassName().startswith("org.python."):
                            # this is a heuristic to eliminate the jython portion of
                            # the java stack trace
                            if not opts.javatrace:
                                break
                        jlines.append("  %s\n" % frame.toString())
                        if frame.getClassName().startswith("java.lang.reflect"):
                            reflect = True
                        else:
                            reflect = False
                    lines.extend(jlines)
                output += indent("".join(lines)).rstrip()
            return output

    def get_exception_type(self):
        if self.exception:
            return self.exception[0]
        else:
            return None

    def get_exception_message(self):
        if self.exception:
            return self.exception[1]
        else:
            return None


ST_WIDTH = 8


def run_test(name, test, config):
    patterns = filter.patterns
    level = root.level
    filter.patterns = config.log_categories
    root.setLevel(config.log_level)

    parts = name.split(".")
    line = None
    output = ""
    for part in parts:
        if line:
            if len(line) + len(part) >= (WIDTH - ST_WIDTH - 1):
                output += "%s. \\\n" % line
                line = "    %s" % part
            else:
                line = "%s.%s" % (line, part)
        else:
            line = part

    if line:
        output += "%s %s" % (line, (((WIDTH - ST_WIDTH) - len(line)) * "."))
    sys.stdout.write(output)
    sys.stdout.flush()
    interceptor.begin()
    start = time.time()
    try:
        runner = test()
    finally:
        interceptor.reset()
    end = time.time()
    if interceptor.dirty:
        if interceptor.last != "\n":
            sys.stdout.write("\n")
        sys.stdout.write(output)
    print(" %s" % colorize_word(runner.status()))
    if runner.failed() or runner.skipped():
        print(runner.get_formatted_exception_trace())
    root.setLevel(level)
    filter.patterns = patterns
    return TestResult(end - start,
                      runner.passed(),
                      runner.skipped(),
                      runner.failed(),
                      runner.get_exception_type(),
                      runner.get_exception_message(),
                      runner.get_formatted_exception_trace())


class TestResult:

    def __init__(self, time, passed, skipped, failed, exception_type, exception_message, formatted_exception_trace):
        self.time = time
        self.passed = passed
        self.skipped = skipped
        self.failed = failed
        self.exception_type = exception_type
        self.exception_message = exception_message
        self.formatted_exception_trace = formatted_exception_trace


class FunctionTest:

    def __init__(self, test):
        self.test = test

    def name(self):
        return "%s.%s" % (self.test.__module__, self.test.__name__)

    def run(self):
        return run_test(self.name(), self._run, config)

    def _run(self):
        runner = Runner()
        runner.run("test", lambda: self.test(config))
        return runner

    def __repr__(self):
        return "FunctionTest(%r)" % self.test


class MethodTest:

    def __init__(self, cls, method):
        self.cls = cls
        self.method = method

    def name(self):
        return "%s.%s.%s" % (self.cls.__module__, self.cls.__name__, self.method)

    def run(self):
        return run_test(self.name(), self._run, config)

    def _run(self):
        runner = Runner()
        inst = self.cls(self.method)
        test = getattr(inst, self.method)

        if hasattr(inst, "configure"):
            runner.run("configure", lambda: inst.configure(config))
            if runner.halt():
                return runner
        if hasattr(inst, "setUp"):
            runner.run("setup", inst.setUp)
            if runner.halt():
                return runner

        runner.run("test", test)

        if hasattr(inst, "tearDown"):
            runner.run("teardown", inst.tearDown)

        return runner

    def __repr__(self):
        return "MethodTest(%r, %r)" % (self.cls, self.method)


class PatternMatcher:

    def __init__(self, *patterns):
        self.patterns = patterns

    def matches(self, name):
        for p in self.patterns:
            if match(name, p):
                return True
        return False


class FunctionScanner(PatternMatcher):

    def inspect(self, obj):
        return type(obj) == types.FunctionType and self.matches(obj.__name__)

    def descend(self, func):
        # the None is required for older versions of python
        return
        yield None

    def extract(self, func):
        yield FunctionTest(func)


class ClassScanner(PatternMatcher):

    def inspect(self, obj):
        return type(obj) in CLASS_TYPES and self.matches(obj.__name__)

    def descend(self, cls):
        # the None is required for older versions of python
        return
        yield None

    def extract(self, cls):
        for name in sorted(dir(cls)):
            obj = getattr(cls, name)
            if hasattr(obj, '__call__') and name.startswith("test"):
                yield MethodTest(cls, name)


class ModuleScanner:

    def inspect(self, obj):
        return type(obj) == types.ModuleType

    def descend(self, obj):
        for name in sorted(dir(obj)):
            yield getattr(obj, name)

    def extract(self, obj):
        # the None is required for older versions of python
        return
        yield None


class Harness:

    def __init__(self):
        self.scanners = [
            ModuleScanner(),
            ClassScanner("*Test", "*Tests", "*TestCase"),
            FunctionScanner("test_*")
        ]
        self.tests = []
        self.scanned = []

    def scan(self, *roots):
        objects = list(roots)

        while objects:
            obj = objects.pop(0)
            for s in self.scanners:
                if s.inspect(obj):
                    self.tests.extend(s.extract(obj))
                    for child in s.descend(obj):
                        if not (child in self.scanned or child in objects):
                            objects.append(child)
            self.scanned.append(obj)


modules = opts.modules
if not modules:
    modules.extend(["proton_tests"])
h = Harness()
for name in modules:
    m = __import__(name, None, None, ["dummy"])
    h.scan(m)

filtered = [t for t in h.tests if is_included(t.name())]
ignored = [t for t in h.tests if is_ignored(t.name())]
total = len(filtered) + len(ignored)

if opts.xml and not list_only:
    xmlr = JunitXmlStyleReporter(opts.xml)
    xmlr.begin()
else:
    xmlr = None


def runthrough():
    passed = 0
    failed = 0
    skipped = 0
    start = time.time()
    for t in filtered:
        if list_only:
            print(t.name())
        else:
            st = t.run()
            if xmlr:
                xmlr.report(t.name(), st)
            if st.passed:
                passed += 1
            elif st.skipped:
                skipped += 1
            elif st.failed:
                failed += 1
                if opts.hoe:
                    break
    end = time.time()

    run = passed + failed

    if not list_only:
        if passed:
            _pass = "pass"
        else:
            _pass = "fail"
        if failed:
            outcome = "fail"
        else:
            outcome = "pass"
        if ignored:
            ign = "ignored"
        else:
            ign = "pass"
        if skipped:
            skip = "skip"
        else:
            skip = "pass"
        sys.stdout.write(colorize("Totals: ", 1))
        totals = [colorize_word("total", "%s tests" % total),
                  colorize_word(_pass, "%s passed" % passed),
                  colorize_word(skip, "%s skipped" % skipped),
                  colorize_word(ign, "%s ignored" % len(ignored)),
                  colorize_word(outcome, "%s failed" % failed)]
        sys.stdout.write(", ".join(totals))
        if opts.hoe and failed > 0:
            print(" -- (halted after %s)" % run)
        else:
            print("")
        if opts.time and run > 0:
            sys.stdout.write(colorize("Timing:", 1))
            timing = [colorize_word("elapsed", "%.2fs elapsed" % (end - start)),
                      colorize_word("average", "%.2fs average" % ((end - start) / run))]
            print(", ".join(timing))

    if xmlr:
        xmlr.end()

    return failed


def main():
    limit = opts.count
    count = 0
    failures = False
    while limit == 0 or count < limit:
        count += 1
        if runthrough():
            failures = True
            if count > 1:
                print(" -- (failures after %s runthroughs)" % count)
        else:
            continue

    if failures:
        return(1)
    else:
        return(0)



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import optparse
from proton import Message, Url
from proton.reactor import Container
from proton.handlers import MessagingHandler, TransactionHandler


class TxSend(MessagingHandler, TransactionHandler):
    def __init__(self, url, messages, batch_size):
        super(TxSend, self).__init__()
        self.url = Url(url)
        self.current_batch = 0
        self.committed = 0
        self.confirmed = 0
        self.total = messages
        self.batch_size = batch_size

    def on_start(self, event):
        self.container = event.container
        self.conn = self.container.connect(self.url)
        self.sender = self.container.create_sender(self.conn, self.url.path)
        self.container.declare_transaction(self.conn, handler=self)
        self.transaction = None

    def on_transaction_declared(self, event):
        self.transaction = event.transaction
        self.send()

    def on_sendable(self, event):
        self.send()

    def send(self):
        while self.transaction and self.sender.credit and (self.committed + self.current_batch) < self.total:
            seq = self.committed + self.current_batch + 1
            msg = Message(id=seq, body={'sequence': seq})
            self.transaction.send(self.sender, msg)
            self.current_batch += 1
            if self.current_batch == self.batch_size:
                self.transaction.commit()
                self.transaction = None

    def on_accepted(self, event):
        if event.sender == self.sender:
            self.confirmed += 1

    def on_transaction_committed(self, event):
        self.committed += self.current_batch
        if self.committed == self.total:
            print("all messages committed")
            event.connection.close()
        else:
            self.current_batch = 0
            self.container.declare_transaction(self.conn, handler=self)

    def on_disconnected(self, event):
        self.current_batch = 0


parser = optparse.OptionParser(usage="usage: %prog [options]",
                               description="Send messages transactionally to the supplied address.")
parser.add_option("-a", "--address", default="localhost:5672/examples",
                  help="address to which messages are sent (default %default)")
parser.add_option("-m", "--messages", type="int", default=100,
                  help="number of messages to send (default %default)")
parser.add_option("-b", "--batch-size", type="int", default=10,
                  help="number of messages in each transaction (default %default)")
opts, args = parser.parse_args()

try:
    Container(TxSend(opts.address, opts.messages, opts.batch_size)).run()
except KeyboardInterrupt:
    pass
